
package softparsmap;


import java.util.*;
import java.lang.*;

/**
 * This child iterator will take into account that some of a the
 * edges in a gene tree might be collapsed and therefore skip
 * those nodes. Using this iterator and while flagging nodes collapsed
 * provides a simple way to temporary collapse nodes in a gene tree.
 */
public class ChildIterator implements Iterator {

    // The list which will hold the nodes.
    private List nodes = null;

    // GeneNode list pointer.
    private int nodesPoi = 0;

    private boolean storeWeakNodes;

    private List weakNodes = null;

    /**
     * This iterator needs a gene node. 
     *
     * @param node is a gene node
     */
    public ChildIterator(GeneNode node) {
	this(node, false);
    }

    /**
     * This iterator needs a gene node and it is possible to store
     * weak edges that are skipped when stepping through the
     * children of the gene node.
     *
     * @param node is the gene node
     * @param storeWeakNodes specifies of weak edges should be stored
     */
    public ChildIterator(GeneNode node, boolean storeWeakNodes) {
	this.nodes = new Vector();
	this.storeWeakNodes = storeWeakNodes;
	if(this.storeWeakNodes) {
	    this.weakNodes = new Vector();
	}
	if(node.hasChildren()) {
	    addAll(node.getChildren());
	}
    }
    
    /**
     *  See Iterator.java
     *
     *  @see java.util.Iterator
     */
    public boolean hasNext() {
	return nodesPoi<nodes.size();
    }
    
    /**
     *  See Iterator.java
     *
     *  @see java.util.Iterator     
     */	
    public Object next() {
	if(hasNext()) {
	    GeneNode n = (GeneNode)this.nodes.get(this.nodesPoi);
	    this.nodesPoi++;
	    while( n!=null && n.getCollapsedEdge() ) {
		if(n.hasChildren()) {
		    addAll(n.getChildren());
		}
		if(hasNext()) {
		    n = (GeneNode)this.nodes.get(nodesPoi);
		    this.nodesPoi++;
		} else {
		    n = null;
		}
	    }
	    if(n!=null) {
		return n;
	    }
	}
	return null;
    }
    
    private void addAll(Collection c) {
	if(!this.storeWeakNodes) {
	    this.nodes.addAll(c);
	    return;
	}
	for(Iterator i = c.iterator();i.hasNext();) {
	    GeneNode n = (GeneNode)i.next();
	    this.nodes.add(n);
	    if(n.isEdgeWeak() && n.hasChildren()) {
		this.weakNodes.add(n);
	    }
	}
    }

    private List getWeakNodes() {
	if(hasNext() || !this.storeWeakNodes) {
	    throw new FatalException
		("Can only be called if the iterator has reached the end and "+
		 "the 'storeWeakNodes' flag is true.");
	}
	return this.weakNodes;
    }
    
    
    /**
     *  This method can not be used and a FatalException will be
     *  thrown if called.
     *
     *  @throws FatalExcetpion if this method is called.
     */
    public void remove() {
	throw new FatalException("Can not be used.");
    }

}
