
package softparsmap;


import ooc.*;
import java.util.*;
import java.io.*;
import java.text.*;

/**
 * This program is used to compute a task.
 * To get usage of a specific task, type
 * <pre>
 * Compute [property file] [task did]
 * </pre>
 */
public abstract class Compute extends LinkedTagInstance {

    //private PropertyFile _propertyFile = null;

    private List _args = null;

    //public void setupTagInstance(Tag tag, Object propertyFile, Object args) {
    //	super.setupTagInstance(tag);
    //	_propertyFile = (PropertyFile)propertyFile;
    //	_args = (List)args;
    //}

    /**
     * Returns the property file.
     *
     * @return the property file
     * @see PropertyFile
     */    
    public PropertyFile getPropertyFile() {
	return (PropertyFile)getContainer();
    }

    /**
     * Returns a list with remaining arguments.
     *
     * @return the list with remaining arguments.
     */
    public List getArgs() {
	return _args;
    }

    public void setArgs(List args) {
	_args = args;
    }

    /**
     * Returns the number of remaining arguments.
     *
     * @return the number of remianing arguments.
     */
    public int numberArgs() {
	return getArgs().size();
    }

    /**
     * Returns the specified argument.
     *
     * @return the specified argument.
     */
    public String getArg(int i) {
	return (String)getArgs().get(i);
    }

    /**
     * If this methods returns true, {@link Compute} will print
     * header information and more when executed.
     *
     * @return true if extra header information are printed
     */
    public boolean getPrintFamilyInfo() {
	return getAttributeBoolean("print_family_info");
    }
 
    /**
     * Returns true if the data source should be initiated.
     *
     * @return true if the data source should be initiated.
     */
    public boolean initiateDataSource() {
	return getAttributeBoolean("initiate_data_source");
    }
    
    /**
     * Is called by main method to run task.
     */
    public abstract void computeTask();
    
    /**
     * Returns a file given the family number and the template attribute name.
     *
     * @return a file given the family number and the template attribute name.
     */
    public File getTargetFile(String templateAttributeName,
			      Integer familyNumber) {
	StringTemplate st =
	    new StringTemplate(getAttributeString(templateAttributeName),
			       getAttributeString("markers_begin"),
			       getAttributeString("markers_end"));
	st.setMarker(getAttributeString("marker_family_number"),
		     familyNumber.toString());
	//st.paintTemplateWithMarkers();
	String fileName = st.getPaintedTemplate();
	try {
	    File file = new File(fileName);
	    File dir = file.getParentFile();
	    if(dir!=null && !dir.exists()) {
		dir.mkdirs();
	    }
	    if (file.exists()) {
		if (!file.delete()) {
		    throw new FatalException
			("Can not delete existing target file "+ fileName);
		}
	    } else {
		if (!file.createNewFile()) {
		    throw new FatalException
			("Can not create targete file "+fileName);
		}
	    }
	    return file;
	} catch (IOException e) {
	    throw new FatalException("Can not get target file "+fileName, e);
	}
    }

    /**
     * Returns a singleton instance of {@link Inparalogous} definied by
     * attribute 'inparalogous' which this task can use
     * regarding inparalogous.
     *
     * @return a inparalogous instance
     * @see Inparalogous
     */
    public Inparalogous getInparalogous() {
	return (Inparalogous)getAttribute("inparalogous");
    }
    
    /**
     * Returns a parser defined by attribute 'attributeName'.
     *
     * @param attributeName is the name of the attribute that contain
     *        the parser did
     * @see TreeParser
     */
    public TreeParser getTreeParser(String attributeName) {
	return (TreeParser)getAttribute(attributeName);
    }

    /**
     * Returns true if the task is using a dotted progress indicator.
     *
     * @return true if the task is using a dotted progress indicator.
     */
    public boolean useProgressIndicator() {
	return getAttributeBoolean("use_progress_indicator");
    }
    
    protected int progress(int now, int loop) {
	if(now>loop) {
	    System.out.print(".");
	    now = 0;
	}
	return now+1;
    }
    
    /**
     * The main method called to do a task.
     */
    public static void main(String[] args) throws Exception {

	// Checking arguments
	if(args.length<2) {
	    System.out.println("Usage: Compute [property file] [task did]");
	    return;
	}
	String propertyFileName = args[0];
	String taskDid = args[1];
	List argList = new Vector();
	for (int i=2;i<args.length;i++) {
	    argList.add(args[i]);
	}
		
	// creating property file
	System.out.println("Parsing property file ...");
	PropertyFile propertyFile =
	    new PropertyFile(new File(propertyFileName));

	// creating task
	System.out.println("Initiating task ...");
	Compute task = (Compute)propertyFile.getTag("task::"+taskDid)
	    .getTagInstance();
	task.setArgs(argList);
	task.computeTask();
	
    }	

    
    
}
