
package softparsmap;


import java.util.*;
import java.io.*;
import java.text.*;

/**
 * This program is used when computing a task on a specified family group.
 * If used without any arguments, the usage will be printed. Usage:
 * <pre>
 * Compute [property file] [task did] [family group did] [ | [family #]]
 * </pre>
 * If a family number is added as the last argument, this family is
 * the only family that will be computed and it will be done in verbose
 * mode.
 */
public abstract class ComputeFamilyGroup extends Compute {
    
    public void computeTask() {
	
	// Checking arguments
	if(!(numberArgs()==1 || numberArgs()==2)) {
	    System.out.println("Usage: Compute [property file] [task did] "
			       +"[family group did] [ | [family #]]");
	    return;
	}
	String groupDid = getArg(0);
	int familyNumber = -1;
	if(numberArgs()==2) {
	    familyNumber = Integer.parseInt(getArg(1));
	}
	
	// creating family group
	System.out.println("Initiating family group ...");
	FamilyGroup familyGroup = (FamilyGroup)getContainer()
	    .getTag("family_group::"+groupDid).getTagInstance();
	//FamilyGroup familyGroup = getPropertyFile().getFamilyGroup(groupDid);
	Set families = familyGroup.getFamilies();

	// if true, initiating data source
	if(initiateDataSource()) {
	    System.out.println("Initiating data source ...");
	    familyGroup.getDataSource().initiate();
	}

	// setup is finished
	System.out.println("Computing task\n");
	
	// run before task
	beforeCompute(familyNumber>0);
	
	// Building group of families or a single family
	for(Iterator i = families.iterator();i.hasNext();) {
	    // Creating the family
	    Family family = (Family)i.next();
	    try {
		// building family
		if (familyNumber>0) {
		    if(family.getFamilyNumber().intValue()==familyNumber) {
			computeTaskHelp(family, true);
		    }
		} else {
		    computeTaskHelp(family, false);
		}
	    } catch (Throwable e) {
		System.out.println("\nFor family "+family.getFamilyNumber()+
				   " following exception was thrown.");
		e.printStackTrace(System.out);
		System.out.println();
	    }
	}

	// run after task
	afterCompute(familyNumber>0);
	
    }
    private void computeTaskHelp(Family family, boolean verbose)
	throws Throwable {
	if (getPrintFamilyInfo()) {
	    System.out.println
		("Family " + family.getFamilyNumber() + ", started " +
		 new Date().toString()+", "+family.getNumberOfSequences()+
		 " sequences.");
	}
	computeFamily(family, verbose);
	if (getPrintFamilyInfo()) {
	    System.out.println();
	}
    }

    
    /**
     * This method is called before any family from the family group
     * is computed.
     *
     * @param verbose is true if extra information is printed
     */
    public void beforeCompute(boolean verbose) {
    }
 
    /**
     * Compute a family.
     *
     * @param family is the family that will be computed
     * @param verbose is true if extra information is printed
     * @throws Throwable because exceptions are logged for every family
     *         but the task will continue on the rest of the family group.
     */
    public abstract void computeFamily(Family family, boolean verbose)
	throws Throwable;
 
    /**
     * This method is called after all families from the family group
     * have been computed.
     *
     * @param verbose is true if extra information is printed
     */
    public void afterCompute(boolean verbose) {
    }
    
}
