package softparsmap;


import ooc.*;
import java.util.*;
import java.io.*;

/**
 * This task will map species from the species tree to the
 * the nodes of the gene tree for all families in a given family group.
 * The result will be written to one file per family. 
 */
public class ComputeMapInfo extends ComputeFamilyGroup {
    
    public void computeFamily(Family family, boolean verbose)
	throws Throwable {

	// creating gene tree
	GeneNode geneTree = family.getGeneTree();

	// check so that gene tree is binary
	if(!geneTree.isBinaryTree()) {
	    throw new FatalException("Gene input tree is not binary.");
	}
	
	// creating species tree
	SpeciesNode tolTree = family.getFamilyGroup().getDataSource()
	    .getSpeciesTree(geneTree);
	
	// inferring duplication and gene loss
	Mutation mutation = tolTree.inferMutation
	    (geneTree, false, getInparalogous());

	// output tree is ready
	GeneNode outputTree = geneTree;
  
	// adding extra information
	Set nodes = outputTree.getNodes();
	nodes.remove(outputTree.getRoot());
	for (Iterator i = nodes.iterator();i.hasNext();) {
	    Node node = (Node)i.next();
	    // adding family
	    put(node, "family_id", family.getFamilyNumber().toString());
	    // adding parent
	    put(node, "parent_label",node.getParent().getMap().get("label"));
	    // child number
	    int childId = 0;
	    for (Iterator k = node.getParent().getChildren().iterator()
		     ;k!=null && k.hasNext();) {
		childId++;
		if (node.equals(k.next())) {
		    k = null;
		}
	    }
	    put(node, "child_number", new Integer(childId).toString());
	    // method
	    put(node, "method", "1");
	}

	// print result if verbose
	if (verbose) {
	    System.out.println("Species tree: \n"+tolTree.toStringAll());
	    System.out.println("Gene tree: \n"+outputTree.toStringAll());
	    System.out.println("Output:");
	}

	// writing to file
	boolean nullValueFound = false;
	FileWriter fw = new FileWriter
	    (getTargetFile("template_target", family.getFamilyNumber()) );
	String keys = getAttributeString("output_keys");
	for (Iterator i = nodes.iterator();i.hasNext();) {
	    Node node = (Node)i.next();
	    List tagIdPath = getTaxIdPath(node, family);
	    if (tagIdPath.size()==1) {
		if (writeRow(fw, node, keys, (Integer)tagIdPath.get(0),
			     (Integer)tagIdPath.get(0), verbose)) {
		    nullValueFound = true;
		}
	    } else {
		for (int m = 0;m<tagIdPath.size()-1;m++) {
		    int n = m + 1;
		    //for (int n = m+1;n<tagIdPath.size();n++) {
		    if (writeRow(fw, node, keys, 
				     (Integer)tagIdPath.get(n),
				     (Integer)tagIdPath.get(m), verbose)
			    ) {
			    nullValueFound = true;
			}
			//}
		}
	    }
	    //fw.write("\n");
	}
	fw.close();
	
	// if null value found a varning is printed
	if (nullValueFound && getPrintFamilyInfo()) {
	    System.out.println
		("- Warning, one or more null values were found.");
	}
	
    }
    
    private boolean writeRow(FileWriter fw, Node node, String keys, 
			     Integer tagUpper, Integer tagLower, 
			     boolean verbose) 
	throws Throwable {
	StringTokenizer st = new StringTokenizer(keys, ",");
	boolean nullValueFound = false;
	while (st.hasMoreTokens()) {
	    String key = st.nextToken().trim();
	    String value = null;
	    if (key.equals("tax_upper")) {
		value = tagUpper.toString();
	    } else if (key.equals("tax_lower")) {
		value = tagLower.toString();
	    } else {
		value =  (String)node.getMap().get(key);
	    }
	    if (value==null) {
		if (verbose) {
		    System.out.print("[null]");
		}
		fw.write("[null]");
		nullValueFound = true;
	    } else {
		if (verbose) {
		    System.out.print(value);
		}
		fw.write(value);
	    }
	    if (verbose) {
		System.out.print("\t");
	    }
	    fw.write("\t");
	}
	if (verbose) {
	    System.out.print("\n");
	}
	fw.write("\n");	
	return nullValueFound;
    }

    private List getTaxIdPath(Node node, Family family) {
	DataSourceXmlNcbiTaxonomy ds = (DataSourceXmlNcbiTaxonomy)
	    family.getFamilyGroup().getDataSource();
	Map childToParent = ds.getIndex().childToParent;
	Integer nodeTagId = ds.getNcbiTagId(((GeneNode)node).getM());
	Integer childTagId = nodeTagId;
	Integer parentTagId = ds.getNcbiTagId
	    (((GeneNode)node.getParent()).getM());
	List ids = new Vector();
	ids.add(childTagId);
	while (!childTagId.equals(parentTagId)) {
	    childTagId = (Integer)childToParent.get(childTagId);
	    if (childTagId==null) {
		throw new FatalException
		    ("Could not find a path between tax id "+nodeTagId+
		     " and "+parentTagId+". Path found was "+ids+".");
	    }
	    ids.add(childTagId);
	}
	return ids;
    }

    private void put(Node node, String key, Object value) {
	if (value==null) {
	    node.getMap().put(key, "[null]");
	} else {
	    node.getMap().put(key, value);
	}
    }
  
}
