
package softparsmap;


import ooc.*;
import java.util.*;
import java.io.*;
import java.text.*;

/**
 * This task will root unrooted gene trees by minimizing duplications
 * and loss. In-paralogous may also be removed.
 * The result will be written to one file per family. 
 */
public class ComputeRoot extends ComputeFamilyGroup {

    private TreeParser getTreeParser() {
	return getTreeParser("tree_parser_out");
    }

    public void computeFamily(Family family, boolean verbose)
	throws Throwable {

	// creating gene tree
	GeneNode freeGeneTree = family.getGeneTree();
	
	// creating species tree
	SpeciesNode tolTree = family.getFamilyGroup().getDataSource()
	    .getSpeciesTree(freeGeneTree);

	// printing input trees
	if (verbose) {
	    System.out.println("\nFree gene tree:\n"+
			       freeGeneTree.toStringAll());
	    System.out.println("Species tree:\n"+
			       tolTree.toStringAll());
	}
	
	// getting list over rooted gene trees with the lowest
	// duplications and gene loss
	List minGeneTrees = tolTree.minimizeMutation
	    (freeGeneTree, true, getInparalogous());
	
	// collapse edges that have been flaged for collapsing and
	// adding splits from the species tree
	for(Iterator k = minGeneTrees.iterator();k.hasNext();) {
	    GeneNode gn = (GeneNode)k.next();
	    gn.collapseEdges();
	}
	
	// choose the preferred tree.
	PreferredTree comp = (PreferredTree)getAttribute("preferred_tree");
	comp.setFreeGeneTree(freeGeneTree);
	comp.setSpeciesTree(tolTree);
	SortedCollection sc = new SortedCollection(comp);
	sc.addAll(minGeneTrees);
	GeneNode pgt = (GeneNode)sc.first();
	
	// check if the first two candidates are equal
	if (getPrintFamilyInfo()) {
	    if (!verbose && sc.size()>1) {
		Iterator i = sc.iterator();
		if(comp.compare(i.next(),i.next())==0) {
		    System.out.println
			("- There are more then one candidate for the output "+
			 "tree.");
		}
	    }
	}

	// if true, printing the list of trees.
	if (verbose) {
	    System.out.println("Choosing the first tree from following "+
			       "list of preferred tree descriptions: ");
	    for (Iterator i = sc.iterator();i.hasNext();) {
		System.out.println("  "+comp.getDesc(((GeneNode)i.next())));
	    }
	}
	
	// if true print preferred tree
	if (verbose) {
	    System.out.println("Preferred tree: \n"+pgt.toStringAll());
	}

	// if true, resolve uncertainties using the species tree.
	if (getAttributeBoolean("resolve_uncertainties_using_species_splits")){
	    if (verbose) {
		System.out.println
		    ("Resolving uncertainties using the species tree ...");
	    }
	    pgt.insertSpeciesTreeSplits();
	}
	
	// if true, resolve uncertainties using out-groups
	if(getAttributeBoolean("resolve_uncertainties_using_outgroups")) {
	    if(verbose) {
		System.out.println
		    ("Resolving uncertainties using outgroups ...");
	    }
	    pgt.resolveUncertainties(freeGeneTree);
	}
		
	// if true, removing in-paralogous
	Inparalogous inp = getInparalogous();
	if(inp.removeBeforeSaving()) {
	    if (verbose) {
		System.out.println("Removing in-paralogous ... ");
	    }
	    // remove inparalogues
	    pgt = inp.removeInparalogous(family, pgt);
	}

	// if true, resolve uncertainties again after in-paralogous
	if(getAttributeBoolean
	   ("resolve_uncertainties_using_outgroups_before_save")) {
	    if(verbose) {
		System.out.println
		    ("Resolving uncertainties using outgroups ...");
	    }
	    pgt.resolveUncertainties(freeGeneTree);
	}
	
	// if true, print the final tree
	if (verbose) {
	    System.out.println("\nFinal rooted tree \n"+pgt.toStringAll());
	}
	
	// write info. about this build to standard output
	if (getPrintFamilyInfo() && !verbose) {
	    if (!pgt.isBinaryTree()) {
		System.out.println("- Output tree is not binary.");
	    }
	}

	// writing to file
	FileOutputStream fos = new FileOutputStream
	    (getTargetFile("template_target", family.getFamilyNumber()));
	getTreeParser().parseOut(fos, pgt);
	fos.close();

	// if not binary, separate binary part from non binary print to
	// separat file
	if (!pgt.isBinaryTree()) {
	    // get max number on leaf labels
	    int maxLabel = -1;
	    for(Iterator i = pgt.getLeaves().iterator();i.hasNext();) {
		GeneLeaf l = (GeneLeaf)i.next();
		if(maxLabel < l.getLabel()) {
		    maxLabel = l.getLabel();
		}
	    }
	    int nextLeafLabel = maxLabel+100;
	    // find non binary subtrees
	    List roots = new Vector();
	    List numbers = new Vector();
	    findNonBinarySubTrees(family.getPropertyFile(), pgt, null,
				  roots, numbers, nextLeafLabel);
	    // set up writer
	    //String path = pf.getTarget().getAbsolutePath()+"/family"+
	    //		fam.number()+"_not_binary.tree";
	    File fw = getTargetFile("template_target_non_binary",
				    family.getFamilyNumber());
	    Writer writer = new CharsetFileWriter(fw);
	    // write result to file
	    writer.write("not binary! \n");
	    for(int i = 0;i<roots.size();i++) {
		Integer number = (Integer)numbers.get(i);
		GeneNode node = (GeneNode)roots.get(i);
		String out = getTreeParser().parseOut(node);
		if(number==null) {
		    writer.write(out+"\n");
		} else {
		    writer.write(number.toString()+" : "+out+"\n");
		}
	    }
	    writer.close();
	}
    }

    private int findNonBinarySubTrees(PropertyFile pf, GeneNode node,
				      GeneNode parent, List roots,
				      List numbers, int nextLeafLabel) {
	// if it has no children, do nothing
	if(!node.hasChildren()) {
	    return nextLeafLabel;
	}
	//
	if(parent==null) {
	    if(node.numChildren()>2) {
		numbers.add(null);
		roots.add(new GeneLeaf(pf.getEdgeTypeUnknown(),nextLeafLabel));
		numbers.add(new Integer(nextLeafLabel));
		roots.add(node);
		nextLeafLabel++;
	    } else {
		roots.add(node);
		numbers.add(null);
	    }
	} else if(node.numChildren()>2) {
	    // move node to roots
	    parent.removeChild(node);
	    roots.add(node);
	    // add a new leaf instead of this sub tree
	    parent.addChild(new GeneLeaf(pf.getEdgeTypeUnknown(),
					 nextLeafLabel));
	    numbers.add(new Integer(nextLeafLabel));
	    nextLeafLabel++;
	}
	// sending on recursive
	for(Iterator i = new Vector(node.getChildren()).iterator()
		;i.hasNext();) {
	    nextLeafLabel = findNonBinarySubTrees
		(pf, (GeneNode)i.next(), node, roots, numbers,nextLeafLabel);
	}
	// return next leaf number
	return nextLeafLabel;
    }

}
