
package softparsmap;


import java.util.*;
import java.io.*;

/**
 * This data source can be used when the sequence database is an XML
 * document. NCBI Taxonomy is used to extract species tree information.
 * It is necessary to construct an index file before using it in
 * different tasks. Run it without arguments to see usage.
 * See the <a href="{@docRoot}/softparsmap/def.xml">def.xml</a>
 * file for more information. 
 */
public class DataSourceXmlNcbiTaxonomy extends AbstractDataSource {
    
    private DataSourceIndexXmlNcbiTaxonomy index = null;
    
    private File ncbiNamesFile = null;

    private File ncbiNodesFile = null;

    private File databaseFile = null;
    
    private File indexFile = null;
    
    protected File getNcbiNamesFile() {
	if(this.ncbiNamesFile==null) {
	    this.ncbiNamesFile =
		new File(getTag().getAttributeString("ncbi_taxonomy_names_file"));
	}
	return this.ncbiNamesFile;
    }
           
    protected File getNcbiNodesFile() {
	if(this.ncbiNodesFile==null) {
	    this.ncbiNodesFile =
		new File(getAttributeString("ncbi_taxonomy_nodes_file"));
	}
	return this.ncbiNodesFile;
    }

    protected File getDatabaseFile() {
	if(this.databaseFile==null) {
	    this.databaseFile =
		new File(getAttributeString("xml_database_file"));
	}
	return this.databaseFile;
    }

    protected File getIndexFile() {
	if(this.indexFile==null) {
	    this.indexFile =
		new File(getAttributeString("index_file"));
	}
	return this.indexFile;
    }

    public void initiate() {
	getIndex();
    }
    
    public String getSpeciesName(SpeciesNode species) {
	return getNcbiSpeciesName(species);
    }

    public SpeciesNode getSpeciesTree(GeneNode geneTree) {
	return getSpeciesTree(geneTree.getLeaves());
    }
    
    private SpeciesNode getSpeciesTree(Set geneSequences) {
	// makes a copy of all gene sequenses.
	GeneLeaf l;
	SortedSet spec = new TreeSet();
	for(Iterator i = geneSequences.iterator();i.hasNext();) {
	    l = (GeneLeaf)((GeneLeaf)i.next()).cloneNode();
	    Integer speciesTaxId = (Integer)
		getIndex().databaseIdToNodesId.get(new Integer(l.getLabel()));
	    if(speciesTaxId==null) {
		throw new FatalException
		    ("Can not find the sequence with database id="+
		     l.getLabel()+" in the map from the XML database to the "+
		     "id numbers in nodes.dmp.");
	    }
	    SpeciesLeaf ls = new SpeciesLeaf(this, speciesTaxId.intValue());
	    if(spec.contains(ls)) {
		ls = (SpeciesLeaf)spec.tailSet(ls).first();
	    } else {
		spec.add(ls);
	    }
	    ls.addSequence(l);
	}
	// building a tree from all gene sequenses
	TreeSet nodesTree = new TreeSet();
	Node node,nodeParent;
	int nodeTaxId = 0;
	for(Iterator i=spec.iterator();i.hasNext();) {
	    node = (SpeciesLeaf)i.next();
	    nodeTaxId = node.getLabel();
	    while(node!=null && nodeTaxId!=1) {
		Integer parentTaxId = (Integer)
		    getIndex().childToParent.get(new Integer(nodeTaxId));
		nodeParent = new SpeciesNode(this, -1*parentTaxId.intValue());
		if(nodesTree.contains(nodeParent)) {
		    nodeParent = (Node)nodesTree
			.tailSet(nodeParent).first();
		    nodeParent.addChild(node);
		    node = null;
		} else {
		    nodesTree.add(nodeParent);
		    nodeParent.addChild(node);
		    node = nodeParent;
		    nodeTaxId = -1*node.getLabel();
		}
	    }
	}
	Node root = ((Node)nodesTree.first()).getRoot();
	root = (SpeciesNode)root.pruneOneChildNodes();
	root.setParent(null);
	return (SpeciesNode)root;
    }

    public SequenceData getSequenceData(GeneLeaf leaf) {
	try {
	    RandomAccessFile rf =
		new RandomAccessFile(getDatabaseFile(),"r");
	    FilePointerPart fp = (FilePointerPart)getIndex()
		.databaseIdToFilePointerPart.get(new Integer(leaf.getLabel()));
	    if(fp==null) {
		throw new FatalException
		    ("The index file is wrong. Regenerate it.");
	    }
	    AbstractSequenceData ads = (AbstractSequenceData)
		getAttribute("abstract_sequence_data");
	    ads.parseSequence(rf, fp);
	    rf.close();
	    return ads;
	} catch(IOException e) {
	    throw new FatalException(e);
	}
    }
    
    /* NCBI methods */

    public Integer getNcbiTagId(SpeciesNode n) {
	int label = n.getLabel();
	if (label<0) {
	    label = -1*label;
	}
	return new Integer(label);
    }

    public String getNcbiNamesColumn(SpeciesNode n, int column) {
	String row = getNcbiNamesRow(n);
	int i = 1;
	int p0 = 0;
	int p1 = row.indexOf('|');
	while (i<column) {
	    p0 = p1;
	    p1 = row.indexOf('|', p0);
	    i++;
	}
	return row.substring(p0+1,p1);
    }
    
    private String getNcbiNamesRow(SpeciesNode n) {
	try {
	    RandomAccessFile rf =
		new RandomAccessFile(getNcbiNamesFile(),"r");
	    int p = Math.abs(n.getLabel());
	    Long fp = (Long)
		getIndex().nodesIdToNamesFilePointer.get(new Integer(p));
	    if(fp==null) {
		throw new FatalException
		    ("The index file is wrong. Regenerate it.");
	    }
	    rf.seek(fp.longValue());
	    String ret = rf.readLine();
	    rf.close();
	    return ret;
	} catch(IOException e) {
	    throw new FatalException(e);
	}
    }
    
    private String getNcbiSpeciesName(SpeciesNode n) {
	String info = getNcbiNamesRow(n);
	int p0 = info.indexOf("|");
	int p1 = info.indexOf("|",p0+1);
	return info.substring(p0+1,p1).trim();
    }

    /*
    public String getNcbiNodesRow(SpeciesNode n) {
	try {
	    RandomAccessFile rf =
		new RandomAccessFile(getNcbiNodesFile(),"r");
	    int p = Math.abs(n.label());
	    Long fp = (Long)
		getIndex().nodesIdToNodesFilePointer.get(new Integer(p));
	    if(fp==null) {
		throw new FatalException
		    ("The index file is wrong. Regenerate it.");
	    }
	    rf.seek(fp.longValue());
	    String ret = rf.readLine();
	    rf.close();
	    return ret;
	} catch(IOException e) {
	    throw new FatalException(e);
	}
    }
    public String getNcbiRank(SpeciesNode n) throws IOException {
	String info = getNcbiNodesRow(n);
	int p0 = info.indexOf("|");
	int p1 = info.indexOf("|",p0+1);
	int p2 = info.indexOf("|",p1+1);
	return info.substring(p1+1,p2).trim();
    }  
    */



    
    /*****  Index methods *****/

    protected DataSourceIndexXmlNcbiTaxonomy getIndex() {
	if(index!=null) {
	    return this.index;
	}
	// try to load index file
	try {
	    FileInputStream fis = new FileInputStream(getIndexFile());
	    ObjectInputStream ois = new ObjectInputStream(fis);
	    this.index = (DataSourceIndexXmlNcbiTaxonomy)ois.readObject();
	    ois.close();
	    fis.close();
	    //
	    return this.index;
	} catch(IOException e) {
	    throw new FatalException(e);
	} catch(ClassNotFoundException e) {
	    throw new FatalException(e);
	}
    }

    private void setIndex(DataSourceIndexXmlNcbiTaxonomy index) {
	this.index = index;
    }    
    
    /**
     *  Will build a new index file. 
     */
    private void buildAndSaveIndex() throws IOException, FatalException {
	//
	System.out.println("Start building index file for data source "+
			   getTagRef());
	
	// create index class
	setIndex(new DataSourceIndexXmlNcbiTaxonomy());

	// parsing NCBI nodes and creating index
	parseNcbiNodes();
	
	// 
	getNodeNumbers();

	// purging the index file
	purgeIndex();
	
	// save index to file for future use
	System.out.println("Saving index to file");
	if(getIndexFile().exists()) {
	    // remove old index file
	    getIndexFile().delete();
	}
	getIndexFile().createNewFile();
	FileOutputStream fos = new FileOutputStream(getIndexFile());
	ObjectOutputStream oos = new ObjectOutputStream(fos);
	oos.writeObject(getIndex());
	oos.flush();
	oos.close();
	fos.close();
    }
    // Help method for the buildAndSaveIndex method.
    private void purgeIndex() {
	System.out.print("Purging index ");
	Set idNodes = new HashSet();
	//System.out.println("*** databaseIdToNodesId.values="+
	//	   getIndex().databaseIdToNodesId.values());
	for(Iterator i = getIndex().databaseIdToNodesId.values().iterator()
		;i.hasNext();) {
	    Integer id = (Integer)i.next();
	    do {
		idNodes.add(id);
		//System.out.println("* id="+id+", get(id)="+
		//	   getIndex().childToParent.get(id));
		id = (Integer)getIndex().childToParent.get(id);
	    } while(id!=null && id.intValue()!=1);
	}
	//System.out.println("*** idNodes="+idNodes);
	Set notUsedIdNodes =
	    new HashSet(getIndex().nodesIdToNamesFilePointer.keySet());
	notUsedIdNodes.removeAll(idNodes);
	// removing id nodes
	for(Iterator i = notUsedIdNodes.iterator();i.hasNext();) {
	    Integer id = (Integer)i.next();
	    getIndex().childToParent.remove(id);
	    getIndex().nodesIdToNamesFilePointer.remove(id);	    
	}
	System.out.println("");
    }
    
    // Help method for the buildAndSaveIndex method.
    private void parseNcbiNodes() throws IOException {
	int progress = 0;
	// creating the tree with NCBI nodes
	System.out.print("Parsing the NCBI node file ");
	RandomAccessFile rf = new RandomAccessFile(getNcbiNodesFile(),"r");
	int p0,p1;
	Integer child, parent;
	//Vector taxIdParent = new Vector();
	long fp = rf.getFilePointer();
	String line = rf.readLine();
	while(line!=null) {
	    progress = progress(progress, 10000);
	    // pointers
	    p0 = line.indexOf("|");
	    p1 = line.indexOf("|",p0+1);
	    child = new Integer(line.substring(0,p0).trim());
	    parent = new Integer(line.substring(p0+1,p1).trim());
	    // tree of life
	    getIndex().childToParent.put(child, parent);
	    // file pointer
	    //getIndex().nodesIdToNodesFilePointer.put(child, new Long(fp));
	    // next row
	    fp = rf.getFilePointer();
	    line = rf.readLine();
	}
	rf.close();
	System.out.println("");
    }
    // Help method for the buildAndSaveIndex method.
    private void getNodeNumbers() throws IOException {
	int progress = 0;
	// read all names in NCBI
	System.out.print("Parsing the NCBI name file ");
	Vector trans = new Vector();
	Map taxaNames = new Hashtable();
	RandomAccessFile rf = new RandomAccessFile(getNcbiNamesFile(),"r");
	int p0,p1,p2;
	String taxaName;
	Integer taxaId;
	long fp = rf.getFilePointer();
	String line = rf.readLine();
	while(line!=null) {
	    progress = progress(progress, 10000);
	    // parse line
	    p0 = line.indexOf("|");
	    p1 = line.indexOf("|",p0+1);
	    taxaName = line.substring(p0+1,p1).trim().toLowerCase();
	    taxaId = new Integer(line.substring(0,p0).trim());
	    // map the taxa name to tax id
	    taxaNames.put(taxaName,taxaId);
	    // map taxa id to the file pointer
	    getIndex().nodesIdToNamesFilePointer.put(taxaId, new Long(fp));
	    // move to next line
	    fp = rf.getFilePointer();
	    line = rf.readLine();
	}
	rf.close();
	System.out.println("");


	// creating the mapping from the database to the NCBI
	System.out.print("Parsing the database and mapping its sequences to "+
			 "species in NCBI ");
	rf = new RandomAccessFile(getDatabaseFile(),"r");
	Integer leafLabel, nodeLabel;
	Vector missingSequences = new Vector();
	AbstractSequenceData ads = (AbstractSequenceData)
	    getAttribute("abstract_sequence_data");
	ads.setDatabaseFile(rf);
	while(ads.moveNextSequence()) {
	    progress = progress(progress, 1000);
	    // parse species name and sequence id number
	    taxaName = ads.getSpeciesName().toLowerCase();
	    leafLabel = new Integer(ads.getIdNumber());
	    // find org name in NCBI
	    nodeLabel = (Integer)taxaNames.get(taxaName);
	    //System.out.println("* leafLabel="+leafLabel+
	    //	       ", taxaName="+taxaName);
	    if(nodeLabel!=null) {
		// map database id to NCBI tax id
		getIndex().databaseIdToNodesId.put(leafLabel, nodeLabel);
		// map database id to file pointer part
		getIndex().databaseIdToFilePointerPart.put
		    (leafLabel, ads.newFilePointerPartCurrent());
	    } else {
		// missing sequences
		missingSequences.add("- sequence "+leafLabel+
				     " with org name '"+taxaName+"'.");
	    }
	}
	rf.close();
	System.out.println("");
	
	// reporting missing sequences
	if(missingSequences.size()>0) {
	    System.out.println("Follwing sequences are missing in NCBI");
	    for(Iterator i=missingSequences.iterator();i.hasNext();) {
		System.out.println(i.next().toString());
	    }
	}
	
    }
    private int progress(int now, int loop) {
	if(now>loop) {
	    System.out.print(".");
	    now = 0;
	}
	return now+1;
    }
    
    
    public static void main(String[] args) throws Exception {
	if(args.length!=2) {
	    System.out.println("Usage:  DataSourceXmlNcbiTaxonomy "+
			       "[property file] [data source did]");
	    return;
	}
	PropertyFile pf = new PropertyFile(new File(args[0]));
	DataSourceXmlNcbiTaxonomy dst = (DataSourceXmlNcbiTaxonomy)
	    pf.getTag("data_source::"+args[1]).getTagInstance();
	dst.buildAndSaveIndex();
    }
    
    
}
