
package softparsmap;

/**
 * This interface is used to define different properties of an edge
 * in a gene tree. In order to define your own properties, extend
 * the class {@link AbstractEdgeType} and make the necessary changes in
 * your property file.
 */
public interface EdgeType {

    /**
     * Returns the property file. Used in the gene node to access
     * the property file.
     *
     * @return the property file
     * @see PropertyFile
     */
    public PropertyFile getPropertyFile();

    /**
     * Returns true if the edge value for the gene node is weak, else false.
     *
     * @return true if the edge value for the gene node is weak, else false.
     * @param node the gene node
     * @see GeneNode
     */
    public boolean isEdgeWeak(GeneNode node);

    /**
     * Returns the new edge value after splitting the edge value for
     * the gene node. Used when re-rooting gene trees among other things.
     *
     * @param node the gene node
     * @return the new edge value
     * @see GeneNode
     */
    public double splitEdge(GeneNode node);

    /**
     * Returns the edge value retained when joining the edges of
     * the two gene nodes.
     *
     * @param node1 the first gene node
     * @param node2 the second gene node
     * @return the edge value as the result of joining the two gene
     *         nodes edge values
     * @see GeneNode
     */
    public double joinEdges(GeneNode node1, GeneNode node2);

    /**
     * Returns the name of the edge type.
     *
     * @return the name of this edge type
     */
    public String getName();

    /**
     * Returns a short name of this edge type. Should not be longer then
     * 2-3 letters long. This method is used in concentrated outputs.
     *
     * @return a short name of the edge type
     */
    public String getShortName();

    public double getDefaultNodeEdge();
    
    public double getDefaultLeafEdge();

}
