
package softparsmap;


import java.io.*;
import java.util.*;

/**
 * This class represents a gene family. Since the number of families
 * can be large, this class will only contain a reference to its
 * family group, its family number, and how many sequences it
 * contains.
 */
public class Family implements Comparable {

    private FamilyGroup familyGroup = null;
    
    private Integer familyNumber = null;

    private int numberOfLeaves = -1;

    /**
     * A family needs to know from which family group it comes from.
     * The family number is used by the family to ask the family group
     * for family specific information.
     *
     * @param familyGroup is the family group this family belongs to
     * @param familyNumber is the number of this family
     * @see FamilyGroup
     */
    public Family(FamilyGroup familyGroup, Integer familyNumber) {
	this.familyGroup = familyGroup;
	this.familyNumber = familyNumber;
    }

    /**
     * Returns the property file.
     *
     * @return the property file
     * @see PropertyFile
     */
    public PropertyFile getPropertyFile() {
	return getFamilyGroup().getPropertyFile();
    }

    /**
     * Returns a copy of this family.
     *
     * @return a new instance with the same family group and family number
     */
    public Family cloneFamily() {
	return new Family(getFamilyGroup(), getFamilyNumber());
    }

    /**
     * Returns the family number.
     *
     * @return the family number
     */
    public Integer getFamilyNumber() {
	return this.familyNumber;
    }

    /**
     * Returns the family group to which this family belongs.
     *
     * @return the family group
     * @see FamilyGroup
     */
    public FamilyGroup getFamilyGroup() {
	return this.familyGroup;
    }

    /**
     * Returns the species tree relevant for this gene family. It is
     * retrieved from the data source. To avoid parsing in the gene
     * tree twice, you should save the gene tree and getting the
     * species tree from the data source directly.
     *
     * @return the species tree
     * @see SpeciesNode
     */
    //public SpeciesNode getSpeciesTree() {
    //	return getFamilyGroup().getDataSource().getSpeciesTree(getGeneTree());
    //}

    /**
     * Returns the gene tree. It gets it from the family group using the
     * the family number.
     *
     * @return the gene tree
     * @see GeneNode
     */
    public GeneNode getGeneTree() {
	GeneNode node = getFamilyGroup().getGeneTree(getFamilyNumber());
	node.setFamilyTree(this);
	return node;
    }

    /**
     * Returns the number of sequences of this family.
     *
     * @return the number of sequences of this family
     */
    public int getNumberOfSequences() {
	if(this.numberOfLeaves<0) {
	    this.numberOfLeaves = getGeneTree().getLeaves().size();
	}
	return this.numberOfLeaves;
    }
    
    /**
     * Compare the family number between this and a second family.
     *
     * @param o is the second famliy
     * @return A integer above zero if this family have a higher
     *         family number then 'o', zero if equal, else below zero
     */
    public int compareTo(Object o) {
	return this.getFamilyNumber().intValue()-
	    ((Family)o).getFamilyNumber().intValue();
    }

    /**
     *  The families are the same if their family numbers are the same.
     *
     *  @param o is the other family.
     *  @return True if the their family numbers are the same.
     */
    public boolean equals(Object o) {
	return this.getFamilyNumber().equals(((Family)o).getFamilyNumber());
    }

    /**
     * Returns the hash code and it has to agree with {@link #equals}.
     *
     * @return the hash code
     */
    public int hashCode() {
	return getFamilyNumber().intValue();
    }
    
    /**
     *  Return the family number in string format.
     *
     *  @return the family number in string format
     */        
    public String toString() {
	return getFamilyNumber().toString();
    }


}


