
package softparsmap;


import ooc.*;
import java.util.*;
import java.io.*;

/**
 * This family group assume that there is one file per family
 * and that the family number can be found in the file name. 
 * Further the parser is used to parse in the gene tree found
 * in the file.
 * See the <a href="{@docRoot}/softparsmap/def.xml">def.xml</a>
 * file for more information.
 */
public class FamilyGroupTreesInFiles extends AbstractFamilyGroup {

    private Set _families = null;

    private Map _numberToFile = null;
  
    protected Map getNumberToFileMap() {
	if(_numberToFile==null) {
	    _numberToFile = new Hashtable();
	}
	return _numberToFile;
    }

    public TreeParser getTreeParserIn() {
	return (TreeParser)getAttribute("tree_parser_in");
    }
    
    public GeneNode getGeneTree(Integer familyNumber) {
	try {
	    File gf = (File)getNumberToFileMap().get(familyNumber);
	    return (GeneNode)getTreeParserIn().parseIn(new FileInputStream(gf));
	} catch(Exception e) {
	    throw new FatalException(e);
	}
    }

    public Set getFamilies() {
	if(_families!=null) {
	    return _families;
	}
	// create families
	Set fams = new TreeSet();
	
	// include directory
	if(hasChildTagByName("include_directory")) {
	    StringTemplate st = new StringTemplate
		(getAttributeString("template_tree_file_name"),
		 getAttributeString("markers_begin"),
		 getAttributeString("markers_end") );
	    String numberMarker = getAttributeString("marker_family_number");
	    if (!st.hasMarker(numberMarker)) {
		throw new FatalException
		    ("In tag "+getTagRef()+" the template in attribute "+
		     "template_tree_file_name does not contain the marker '"+
		     numberMarker+"'.");
	    }
	    for(Iterator it = getChildIteratorByName("include_directory")
		    ;it.hasNext();) {
		Tag idt = (Tag)it.next();
		String dirName = 
		    idt.getAttributeString("tree_files_directory");
		File dir = new File(dirName);
		if(!dir.exists() || !dir.isDirectory()) {
		    throw new FatalException
			("The directory '"+dirName+"' in tag "+idt.getTagRef()+
			 " does not exist.");
		}
		File[] treeFiles = dir.listFiles();	
		for (int i=0;i<treeFiles.length;i++) {
		    File treeFile = treeFiles[i];
		    if (treeFile.isFile() && treeFile.canRead()) {
			try {
			    st.paintTemplate(treeFile.getName());
			    Integer number =
				new Integer(st.getFirstValue(numberMarker));
			    Family newFam = new Family(this, number);
			    getNumberToFileMap().put(number, treeFiles[i]);
			    fams.add(newFam);
			} catch (RuntimeException e) {}
		    }
		}
	    }
	    if (fams.size()==0) {
		throw new FatalException
		    ("There exist include_directory tags under tag "+
		     getTagRef()+" but no tree files matching template '"+
		     getAttributeString("template_tree_file_name")+"'.");
	    }
	}
	
	// include groups
	if(hasChildTagByName("include_group")) {
	    for(Iterator i = getChildIteratorByName("include_group")
		    ;i.hasNext();) {
		Tag t = (Tag)i.next();
		FamilyGroup fm = (FamilyGroup)t.getAttribute("family_group");
		fams.addAll(filterNonValid(t, fm));
	    }
	}
	
	// exclude groups
	if(hasChildTagByName("exclude_group")) {
	    for(Iterator i = getChildIteratorByName("exclude_group")
		    ;i.hasNext();) {
		Tag t = (Tag)i.next();
		FamilyGroup fm = (FamilyGroup)t.getAttribute("family_group");
		fams.removeAll(filterNonValid(t, fm));
	    }
	}
	    
	// save and return set
	_families = fams;
	return _families;
    }

    private Set filterNonValid(Tag ieGroup, FamilyGroup familyGroup) {
	Set families = new TreeSet();
	Set numbers = null;
	if(!ieGroup.getAttributeString("family_numbers")
	   .equalsIgnoreCase("all")) {
	    numbers = new TreeSet();
	    StringTokenizer st = new StringTokenizer
		(ieGroup.getAttributeString("family_numbers"),",");
	    while(st.hasMoreTokens()) {
		numbers.add(new Integer(st.nextToken().trim()));
	    }
	    if(numbers.size()==0) {
		throw new FatalException
		    ("Attribute 'family_numbers' in tag "+ieGroup.getTagRef()+
		     " must contain comma seperated family numbers or 'all'.");
	    }
	}
	//System.out.println("*** tag="+getTag().getTagPath()+
	//		   ", numbers="+numbers); 
	int minFamilyNumber = ieGroup.getAttributeInteger("min_family_number");
	int maxFamilyNumber = ieGroup.getAttributeInteger("max_family_number");
	int minNumberLeaves = ieGroup.getAttributeInteger("min_number_leaves");
	int maxNumberLeaves = ieGroup.getAttributeInteger("max_number_leaves");
	// if valid, copy, change family group and add it
	for(Iterator i = familyGroup.getFamilies().iterator();i.hasNext();) {
	    Family fam = (Family)i.next();
	    if((numbers==null || numbers.contains(fam.getFamilyNumber())) &&
	       minFamilyNumber<=fam.getFamilyNumber().intValue() &&
	       fam.getFamilyNumber().intValue() <= maxFamilyNumber &&
	       minNumberLeaves <= fam.getNumberOfSequences() &&
	       fam.getNumberOfSequences() <= maxNumberLeaves ) {
		fam = fam.cloneFamily();
		//fam.setFamilyGroup(familyGroup);
		families.add(fam);
	    }
	}
	//
	return families;
    }
   
}
