
package softparsmap;

import java.io.*;
import java.util.*;

/**
 *  Defines a gene leaf in a gene tree.
 */
public class GeneLeaf extends GeneNode implements Comparable {

    private int inparalogousOrder = -1;
    
    /**
     *  Construct a gene leaf given an edge type, a label, and an edge value.
     *
     *  @param edgeType is the edge type
     *  @param label is the gene label
     *  @param edge is the value of edge for this leaf
     */
    public GeneLeaf(EdgeType edgeType, int label, double edge) {
	super(edgeType, label, edge);
	//if(label<=0) {
	//    throw new FatalException
	//	("GeneLeaf label(in this case "+label+") can not be below 1.");
	//}
    }

    /**
     *  Same as using the constructor GeneLeaf(edgeType, label, 1.0);
     */        
    public GeneLeaf(EdgeType edgeType, int label) {
	this(edgeType, label, getDefaultEdge(edgeType));
    }
    private static double getDefaultEdge(EdgeType edgeType) {
	if (edgeType==null) {
	    return 1.0;
	}
	return edgeType.getDefaultLeafEdge();
    }

    protected Map getAllInfo() {
	Map info = super.getAllInfo();
	// inparalogous order
	if(this.inparalogousOrder>0) {
	    info.put("IO",new Integer(this.inparalogousOrder).toString());
	}
	return info;
    }

    protected int getInparalogousOrder() {
	return this.inparalogousOrder;
    }

    protected void setInparalogousOrder(int order) {
	this.inparalogousOrder = order;
    }
    
    public SpeciesNode getM() {
	if(this.m==null) {
	    throw new FatalException("The species for which this gene "+
				     "originate has not been set.");
	}
	return this.m;
    }

    /**
     * Sets the species in which this sequence was found.
     *
     * @param m is the species in which this sequence was found
     */
    public void setM(SpeciesNode m) {
    	this.m = m;
    }
    
    protected Node purgeThis() {
	SpeciesNode m = this.m;
	super.purgeThis();
	this.m = m;
    	return this;
    }


    protected GeneNode resetFlags() {
	SpeciesNode m = this.m;
	super.resetFlags();
	this.m = m;
    	return this;
    }
     
    /**
     *  Leaves can not have children so this will throw an
     *  {@link FatalException}.
     *
     *  @throws FatalException since a child can not have children.
     */
    public Set getChildren() {
	throw new FatalException("Leaves can not have children.");
    }

    public GeneNode setCollapsedEdge(boolean collapsedEdge) {
	if(collapsedEdge) {
	    throw new FatalException("Can not collapse an edge of a leaf.");
	}
    	super.setCollapsedEdge(collapsedEdge);
    	return this;
    }

    protected Node setLabel(int label) {
	if(label<=0) {
	    throw new FatalException("GeneLeaf label can not be negative.");
	}
	this.label = label;
	return this;
    }
    
    public Node pruneOneChildNodes() {
	return this;
    }
   
    public GeneNode cloneNode() {
	GeneLeaf nl =  new GeneLeaf(getEdgeType(), getLabel(), getEdgeValue());
	nl.inparalogousOrder = this.inparalogousOrder;
	cloneNodePaintAttributes(nl);
	return nl;
    }
       
    public boolean rootedEquals(GeneNode leaf) {
    	return this.equals(leaf);
    }
    
    public GeneLeaf findLeaf(GeneLeaf leaf) {
	if(this.equals(leaf)) {
	    return this;
	} else {
	    return null;
	}
    }
    
    public boolean isBinaryTree() {
	return true;
    }
    
    public int compareTo(Object o) {
	return this.getLabel()-((GeneNode)o).getLabel();
    }
    
    public GeneNode rerootThis()  {
	throw new FatalException("Leaves can not have children.");
    }
    
    public String toString() {
	return new Integer(getLabel()).toString();
    }
    
    public void setNewLabel() {
	this.setLabel(nextUnique());
    }
    
    /**
     *  Returns a new leaf with an unique label.
     *
     *  @return a new leaf with an unique label. 
     */
    public static GeneLeaf nextLeaf(EdgeType edgeType) {
	return new GeneLeaf(edgeType, nextUnique());
    }

}
