
package softparsmap;


import java.io.*;
import java.util.*;
import java.lang.*;

/**
 * Defines a gene node in a gene tree.
 */
public class GeneNode extends Node {

    private double edge;

    private int duplications = -1;

    private int geneLosses = -1;

    protected SpeciesNode m = null;

    private Set mm = null;

    private Set sl = null;
    
    private boolean collapsedEdge = false;

    private EdgeType edgeType = null;

    private Family family = null;
    
    /**
     * Constructs a gene node given an edge type and a label.
     *
     * @param edgeType is the edge type that the edge of this node belong to.
     * @param label is the label of this node
     * @see EdgeType
     */
    public GeneNode(EdgeType edgeType, int label) {
	this(edgeType, label, getDefaultEdge(edgeType));
    }
    private static double getDefaultEdge(EdgeType edgeType) {
	if (edgeType==null) {
	    return 1.0;
	}
	return edgeType.getDefaultNodeEdge();
    }

    /**
     * Constructs a gene node given an edge type, edge value, and label.
     *
     * @param edgeType is the edge type that the edge of this node belong to.
     * @param label is the label of this node
     * @param edge is the edge value of this node
     * @see EdgeType
     */
    public GeneNode(EdgeType edgeType, int label, double edge) {
	super(label);
	this.edge = edge;
	this.setEdgeType(edgeType);
	//if(label>=0) {
	//   throw new FatalException("Internal node label can "+
	//			     "not be positive.");
	//}
    }

    /**
     * Constructor for extending classes. Does nothing.
     */
    protected GeneNode() {
    }

    /**
     *  See Node
     *  @see Node
     */
    protected Node purgeThis() {
	super.purgeThis();
	this.m = null;
	this.mm = null;
	this.sl = null;
    	return this;
    }

    /**
     * Returns the family that this node belongs to.
     *
     * @return the family that this node belongs to
     * @see Family
     */
    public Family getFamily() {
	return this.family;
    }

    /**
     * Sets the family that this node belongs to.
     *
     * @param family is the family that this node belong to
     */
    public void setFamily(Family family) {
	this.family = family;
    }

    /**
     * Sets the family for this subtree.
     *
     * @param family is the family for this subtree
     */
    public void setFamilyTree(Family family) {
	this.setFamily(family);
	if(hasChildren()) {
	    for(Iterator i = getChildren().iterator();i.hasNext();) {
		((GeneNode)i.next()).setFamilyTree(family);
	    }
	}
    }

    /**
     * Returns true if this node has a family.
     *
     * @return true if this node has a family
     */
    public boolean hasFamily() {
	return this.family!=null;
    }
    
    protected Map getAllInfo() {
	Map info = super.getAllInfo();
	//
	if(hasSL()) {
	    info.put("SL(g)",getSL().toString());
	}
	//
	if(hasM()) {
	    info.put("m(g)",getM().toString());
	}
	//
	if(hasMM()) {
	    info.put("M(g)",new TreeSet(getMM()).toString());
	}
	
	// edge info
	try {
	    String ei = getEdgeInfo();
	    if(ei!=null && !ei.equals("")) {
		info.put("N.i.",getEdgeInfo());
	    }
	} catch(Exception e) {
	    info.put("N.i.","Exception");
	}
	
	// edge value
	try {
	    info.put("E.v.",getEdgeValue()+"-"+getEdgeType().getShortName());
	} catch(Exception e) {
	    info.put("E.v.","Exception");
	}
	
	//
	return info;
    }
   
    /**
     * Try to resolve all uncertainties in this rooted gene tree
     * using the original unrooted gene tree. A new tree will be created.
     *
     * This method can only be called on the root of the tree,
     * else a FatalException will be thrown.
     *
     * @param originalTree is the original unrooted gene tree.
     * @return the root of the new tree
     */
    public GeneNode resolveUncertainties(GeneNode originalTree) {
	// make sure the method is called on the root.
	if(!this.isRoot() || !originalTree.isRoot()) {
	    throw new FatalException("This method can only be called on the "+
				     "the root, 'originalTree' has to be the "+
				     "root as well.");
	}
	// make sure that the leaf sets are correct
	if(!originalTree.getLeaves().containsAll(this.getLeaves())) {
	    throw new FatalException("The leaf set of this tree has to be a "+
				     "subset of 'originalTree's leaf set.");
	}
	// make sure that both trees have the same set of
	GeneNode cot = (GeneNode)originalTree.cloneTree().getLeaves()
	    .iterator().next();
	resolveSub1(this, null, cot);
	// return this as the root of the new tree.
	return this;
    }
    private void resolveSub1(GeneNode geneNode, GeneNode geneNodeOut,
			     GeneNode originalTree) {
	// if true, the geneNode can not be resolved.
	//System.out.println("* In 1, geneNode: "+geneNode+
	//  ", geneNodeOut: "+geneNodeOut);
	if(geneNodeOut==null || geneNode.numChildren()==2) {
	    if(geneNode.numChildren()>2) {
		for(Iterator i = geneNode.getChildren().iterator()
			;i.hasNext();) {
		    GeneNode cn = (GeneNode)i.next();
		    //System.out.println("* In 1.01, c1: "+cn+", c2: null");
		    resolveSub1(cn, null, originalTree);
		}
	    } else if(geneNode.hasChildren()) {
		Iterator i = geneNode.getChildren().iterator();
		GeneNode c1 = (GeneNode)i.next();
		GeneNode c2 = (GeneNode)i.next();
		//System.out.println("* In 1.02, c1: "+c1+", c2: "+c2);
		resolveSub1(c1, c2, originalTree);
		resolveSub1(c2, c1, originalTree);
	    }
	    return;
	}
	// geneNode has to have children.
	if(!geneNode.hasChildren()) {
	    return;
	}
	// this node need to be resolved and we start with mid point
	// reroot original tree at the right point.
	//System.out.println("* In 1.1, geneNode: "+geneNode);
	GeneNode rootedOrg = null;
	{
	    Set oog = geneNodeOut.getLeaves();
	    GeneNode bestNode = null;
	    int bestSize = -1;
	    for(Iterator i = originalTree.getRoot().getNodes().iterator();
		i.hasNext();) {
		GeneNode n = (GeneNode)i.next();
		Set nl1 = n.getLeaves();
		Set nl2 = new HashSet(geneNode.getRoot().getLeaves());
		nl2.removeAll(nl1);
				
		if(nl1.containsAll(oog)) {
		    if(bestSize<0 || bestSize>nl1.size()) {
			bestNode = n;
			bestSize = nl1.size();
		    }
		} else if(nl2.containsAll(oog)) {
		    if(bestSize<0 || bestSize>nl2.size()) {
			bestNode = n;
			bestSize = nl2.size();
		    }
		}
	    }
	    rootedOrg = (GeneNode)bestNode.reroot().getRoot();
	}
	// resolve uncertainties in geneNode
	if(resolveSub4(geneNode, rootedOrg, originalTree)) {
	    for(Iterator h = geneNode.getChildren().iterator();h.hasNext();) {
		resolveSub1((GeneNode)h.next(), null, originalTree);
	    }
	}
    }
    // resume at point for in the alg.
    private boolean resolveSub4(GeneNode geneNode, GeneNode rootedOT,
				GeneNode originalTree) {
	GeneNode otn = (GeneNode)rootedOT.findLCA(geneNode.getLeaves());
	//System.out.println("*** \n"+rootedOT.toStringAll());
	//System.out.println("*** otn: "+otn);
	HashSet[] sg = new HashSet[otn.numChildren()];
	GeneNode[] gc = toNodeArray(geneNode.getChildren());
	GeneNode[] otnc = toNodeArray(otn.getChildren());
	for(int i=0;i<gc.length;i++) {
	    int goesTo = -1;
	    for(int k=0;k<otnc.length;k++) {
		if(otnc[k].getLeaves().containsAll(gc[i].getLeaves())) {
		    goesTo = k;
		} else if(MuNode.anyInCommon(otnc[k].getLeaves(),
					     gc[i].getLeaves()) ) {
		    // gc children can not be devided because there are one
		    // or more children of gc which leaf set exists in more
		    // then one of otn's children.
		    // it is not possible split the children of geneNode.
		    // there might be more uncertainties further down into the
		    // tree that need to be resolved.
		    return true;
		}
	    }
	    // if true, there are more information in the unresolved
	    // tree then in the original tree. could not find a child to otn
	    // that contained all leaves of gc[i].
	    if(goesTo<0) {
		// there might be more uncertainties further down into the
		// tree that needs to be resolved.
		return true;
	    }
	    // add gc[i] to the bucket of the chosen child of otn.
	    if(sg[goesTo]==null) {
		sg[goesTo] = new HashSet();
	    }
	    sg[goesTo].add(gc[i]);
	}
	// introduce new internal nodes.
	for(int i=0;i<sg.length;i++) {
	    if(sg[i]!=null) {
		if(sg[i].size()>1) {
		    GeneNode newNode = otnc[i].cloneNode();
		    for(Iterator it = sg[i].iterator();it.hasNext();) {
			GeneNode n = (GeneNode)it.next();
			geneNode.removeChild(n);
			newNode.addChild(n);
		    }
		    geneNode.addChild(newNode);
		    // if true, try to resolve even further.
		    if(sg[i].size()>2) {
			resolveSub4(newNode, rootedOT, originalTree);
		    }
		}
	    }
	}
	// the geneNode has been split at least one time. return true will
	// keep on resolving uncertainties in the sub tree of geneNode
	return true;
    }
    private GeneNode[] toNodeArray(Collection children) {
	GeneNode[] c = new GeneNode[children.size()];
	int k = 0;
	for(Iterator i=children.iterator();i.hasNext();) {
	    c[k] = (GeneNode)i.next();
	    k++;
	}
	return c;
    }


    /**
     * Returns the edge type of the edge that belong to this node.
     *
     * @return the edge type of the edge that belong to this node
     */
    public EdgeType getEdgeType() {
	return this.edgeType;
    }

    /**
     * Returns the property file.
     *
     * @return the property file
     */
    public PropertyFile getPropertyFile() {
	return getEdgeType().getPropertyFile();
    }
    
    /**
     * Returns the gene node that harbor the split that contain
     * the leaf set 'splitLeavesA'.
     *
     * @param splitLeavesA is the first set that defines a split
     * @param splitLeavesB is the second set that defines a split
     * @return a gene node that harbor the split
     * @see GeneLeaf
     */
    public GeneNode findNodeForUnrootedSplit(Set splitLeavesA,
					     Set splitLeavesB) {
	if(!isRoot()) {
	    throw new FatalException
		("This method can only be called on the root.");
	}
	if(splitLeavesA.size()==0 || splitLeavesB.size()==0) {
	    throw new FatalException("The sets have to be non empty.");
	}
	GeneNode n = findNodeForUnrootedSplitHelp(this, getLeaves(),
					      splitLeavesA, splitLeavesB);
	if(n==null) {
	    //System.out.println("*******************\n"+this.toStringAll());
	    throw new FatalException("Could not find split A="+splitLeavesA+
				     ", B="+splitLeavesB);
	}
	return n;
    }
    private GeneNode findNodeForUnrootedSplitHelp(GeneNode n, Set allLeaves,
					      Set splitA, Set splitB) {
	// if true, check this
	if(!n.isRoot()) {
	    Set nSplitA = n.getLeaves();
	    Set nSplitB = new HashSet(allLeaves);
	    nSplitB.removeAll(nSplitA);
	    //
	    //System.out.println("n="+n+", A="+splitA+", B="+splitB+
	    //		       ", nA="+nSplitA+", nB="+nSplitB);
	    if(nSplitA.containsAll(splitA) && nSplitB.containsAll(splitB) ||
	       nSplitA.containsAll(splitB) && nSplitB.containsAll(splitA)) {
		return this;
	    }
	}
	// check children
	if(n.hasChildren()) {
	    for(Iterator i = n.getChildren().iterator();i.hasNext();) {
		GeneNode ni = findNodeForUnrootedSplitHelp
		    ((GeneNode)i.next(), allLeaves, splitA, splitB);
		if(ni!=null) {
		    return ni;
		}
	    }
	}
	// found non
	return null;
    }
    
    
    /**
     * Collapsing all edges in this subtree where the collapsed
     * flag is set.
     */
    public void collapseEdges() {
	//	System.out.println("this: " + this);
	if(hasChildren()) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		((GeneNode)i.next()).collapseEdges();
	    }
	    Object[] children = getChildren().toArray();
	    for(int i=0;i<children.length;i++) {
		GeneNode n = (GeneNode)children[i];
		//System.out.println("* n: " + n + ", coll: " +
		//n.collapsedgetEdgeValue());
		if(n.hasChildren() && n.getCollapsedEdge()) {
		    //System.out.println("* add child: " + n.getChildren());
		    this.removeChild(n);
		    this.addChildren(n.getChildren());
		}
	    }
	}
    }

    
    /**
     * Insert splits from the species tree if it does not increase
     * the minimum number of duplications and gene loss.
     */
    public void insertSpeciesTreeSplits() {
	// insert splits from the minimum mutation tree
	for(Iterator i=this.nonBinaryNodes().iterator();i.hasNext();) {
	    GeneNode g = (GeneNode)i.next();
	    List partition = MuNode.computePartition
		(g.getChildIterator(), MuNode.MAX_NUMBER_OF_PARTITION);
	    MuNode mg = MuNode.buildGeneTree(partition);
	    
	    // disconnect g's children
	    List gc = new Vector(g.getChildren());
	    for (Iterator k = gc.iterator();k.hasNext();) {
		g.removeChild((GeneNode)k.next());
	    }
	    // remove children of mg from mg and add them to g
	    List mgc = new Vector(mg.getChildren());
	    for (Iterator k = mgc.iterator();k.hasNext();) {
		MuNode mgk = (MuNode)k.next();
		mg.removeChild(mgk);
		g.addChild(mgk);
	    }
	    // replace the leaves of g with the linked nodes
	    for (Iterator k = new Vector(g.getLeaves()).iterator();
		 k.hasNext();) {
		MuNode lg = (MuNode)k.next();
		if (!lg.hasLinkedNode()) {
		    throw new FatalException("Must have a linked node.");
		}
		GeneNode lgp = (GeneNode)lg.getParent();
		lgp.removeChild(lg);
		lgp.addChild(lg.getLinkedNode());
	    }
	}
	// remove splits that was inserted above that do not exist in
	// the species tree.
	recInsertSpeciesTreeSplits(this);
	// collapse all edges which is flagged collapsed.
	this.collapseEdges();
    }
    private void recInsertSpeciesTreeSplits(GeneNode n) {
	if(!n.hasChildren()) {
	    return;
	}
	// down first recursion
	for(Iterator i = n.getChildren().iterator();i.hasNext();) {
	    recInsertSpeciesTreeSplits((GeneNode)i.next());
	}
	// set collapsed flag if n is an added node and the split does
	// not exist in the species tree
	if(n instanceof MuNode) {
	    //System.out.println("** in \n"+n.toStringAll());
	    Set chiM = new HashSet(n.getM().getLeaves());
	    chiM.removeAll(n.getSL());
	    if(chiM.size()>0) {
		n.setCollapsedEdge(true);
	    }
	}
    }

    
    /**
     * Reassigns the label of this node to a new unique number.
     *
     * @return this node for chaining     
     */
    public void setNewLabel() {
	this.setLabel(-1*nextUnique());
    }
    
    /**
     * Return all internal nodes in this subtree.
     *
     * @return a set of gene nodes.
     */
    public Set getInternalNodes() {
	return getInternalNodesRec(new HashSet());
    }
    private Set getInternalNodesRec(Set all) {
	if(!(this instanceof GeneLeaf)) {
	    all.add(this);
	    if(hasChildren()) {
		for(Iterator i=getChildren().iterator();i.hasNext();) {
		    all.addAll(((GeneNode)i.next()).getInternalNodesRec(all));
		}
	    }
	}
	return all;
    }

    /**
     *  Reset information that are found when inferring duplications
     *  and gene loss.
     *
     *  @return This node for chaining
     */
    protected GeneNode resetFlags() {
    	this.setDuplicationsNode(0);
	this.setGeneLossesNode(0);
    	this.setCollapsedEdge(false);
	this.m = null;
	this.mm = null;
	return this;
    }

    /**
     *  Run resetFlags() on all nodes in this subtree.
     *
     *  @return This node for chaining
     */
    protected GeneNode resetFlagsDown() {
	this.resetFlags();
	if(hasChildren()) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		((GeneNode)i.next()).resetFlagsDown();
	    }
    	}
    	return this;
    }
  
    /**
     *  Run resetFlags() on all ancestors to this gene node.
     *
     *  @return This node for chaining     
     */
    protected GeneNode resetFlagsUp() {
	GeneNode node = this;
	do {
	    node.resetFlags();
	    node = (GeneNode)node.getParent();
	} while(node!=null);
	return this;
    }

    /**
     * Counting the number of strong edges in this subtree.
     * If this node is a root will NOT be counted.
     *
     * @return the number of strong edges in this subtree.
     */
    public int countStrongEdges() {
	int n = 0;
	GeneNode node;
	for(Iterator i = getNodes().iterator();i.hasNext();) {
	    node = (GeneNode)i.next();
	    if( !node.isRoot() && !node.isEdgeWeak() ) {
		n++;
	    }
	}
	return n;
    }
    
    /**
     * Return all nodes in the subtree where their collapse flag is
     * set to true.
     *
     * @return a set of nodes.
     */
    public Set collapsedEdges() {
    	return collapsedEdgesSub(new HashSet()); //new TreeSet());
    }
    private Set collapsedEdgesSub(Set all) {
    	if(hasChildren()) {
		for(Iterator i=getChildren().iterator();i.hasNext();) {
			((GeneNode)i.next()).collapsedEdgesSub(all);
		}
    		if(getCollapsedEdge()) {
    			all.add(this);
    		}
    	}
    	return all;
    }

    /**
     * Returns a copy of this GeneNode. Does not connect it to
     * any other node.
     *
     * @return The copy.
     */
    public GeneNode cloneNode()  {
	GeneNode n = new GeneNode(this.edgeType, getLabel(),this.edge);
	cloneNodePaintAttributes(n);
	return n;
    }
    protected void cloneNodePaintAttributes(GeneNode clonedNode) {
	clonedNode.family = this.family;
	clonedNode.m = this.m;
	clonedNode.mm = this.mm;
    }

    /**
     * Returns a copy of this subtree.
     *
     * @return the root of the new subtree
     */
    public GeneNode cloneTree() {
    	GeneNode root = cloneNode();
    	if(hasChildren()) {
	    for(Iterator i = getChildren().iterator();i.hasNext();) {
		root.addChildWithoutPurge(((GeneNode)i.next()).cloneTree());
	    }
	}
	return root;
    }
    
    /**
     * Returns the leaf with given label. The search will only
     * be done in this subtree.
     *
     * @param label is a label
     * @return a leaf with given label. Null if not found
     * @see GeneLeaf
     */
    public GeneLeaf findLeaf(int label) {
	return findLeaf(new GeneLeaf(null, label));
    }

    /**
     * Returns the leaf with the same label as given leaf.
     * The search will only be done in this subtree.
     *
     * @param leaf is the leaf
     * @return a leaf that has the same label as the given leaf.
     *         Null if not found.
     * @see GeneLeaf
     */    
    public GeneLeaf findLeaf(GeneLeaf leaf) {
	if(hasChildren()) {
	    GeneLeaf l;
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		l = ((GeneNode)i.next()).findLeaf(leaf);
		if(l!=null) {
		    return l;
		}
	    }
	}
	return null;
    }

    /**
     * Returns the gene node that exists in this subtree and has
     * the same label.
     *
     * @param node the node to look for.
     * @return the gene node that exists in this subtree and has the
     *         same label.
     */
    public GeneNode findNode(GeneNode node) {
	if(this.equals(node)) {
	    return this;
	}
	if(hasChildren()) {
	    GeneNode n;
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		n = ((GeneNode)i.next()).findNode(node);
		if(n!=null) {
		    return n;
		}
	    }
	}
	return null;
    }

    /**
     * Re-rooting the gene tree that this gene tree is a part of,
     * using midpoint re-rooting. It is dependent on the method
     * {@link #rerootThis}.
     *
     * @return the new root of the tree.
     */
    public GeneNode reroot()  {
	if(this.getParent()!=null) {
	    double splitEdge = getEdgeType().splitEdge(this);
	    GeneNode newRoot = GeneNode.nextNode(this.getEdgeType());
	    newRoot.family = this.family;
	    GeneNode parent = (GeneNode)getParent();
	    parent.rerootThis();
	    parent.removeChild(this);
	    newRoot.addChild(this);
	    newRoot.addChild(parent);
	    // adjusting edge properties
	    this.setEdgeValue(splitEdge);
	    parent.setEdgeValue(splitEdge);
	    parent.setEdgeType(this.getEdgeType());
	    return (GeneNode)newRoot.pruneOneChildNodes();
	} else {
	    return this;
	}
    }

    /**
     * Re-rooting the gene tree that this node is a part of. This
     * gene node will be the root when finished.
     *
     * @return the root of the re-rooted tree.
     */
    public GeneNode rerootThis()  {
	if(this.getParent()!=null) {
	    getRoot().purgeDown();
	    ((GeneNode)this.getParent()).recRerootThis(this);
	}
	return this;
    }
    private void recRerootThis(GeneNode child)  {
	if(this.getParent()!=null) {
	    ((GeneNode)this.getParent()).recRerootThis(this);
	}
	// nu r this rot och den ska bli barn
	if(numChildren()<3) {
	    // this mste tas bort annars s fr den bara ett
	    // barn. se Lemma 1.6
    	    Iterator i = getChildren().iterator();
	    GeneNode notChild = (GeneNode)i.next();
	    if(notChild.equals(child)) {
		notChild = (GeneNode)i.next();
	    }
	    child.addChild(notChild);
	    // Ser till att samman slagningen av kanterna blir rtt.
	    notChild.setEdgeValue(getEdgeType().joinEdges(notChild, child));
	    // no parent
	    child.setParent(null);
	} else {
	    // tar bort medskickade barnet frn barnen(subnodes)
	    this.removeChild(child);
	    // lgger till this som barn under medskickat barn
	    child.addChild(this);
	    this.setEdgeValue(child.edge);
	    this.setEdgeType(child.getEdgeType());
	    //System.out.println("********* child: "+
	    //child.setEdgeValue(Double.NaN);
	    child.setParent(null);
	}
    }
  
    /**
     * Returns true if this subtrees have the same rooted topology as
     * the given subtree.
     *
     * @param root is the root of the subtree that this subtree will
     *             be compared against
     * @return true if their rooted topology is identical, else false
     */
    public boolean rootedEquals(GeneNode root) {
	if(root instanceof GeneLeaf) {
	    return false;
	} else {
	    Collection c0 = this.getChildren();
	    Collection c1 = new Vector();
	    c1.addAll(root.getChildren());
	    if(c0.size()!=c1.size()) {
		return false;
	    } else {
		Iterator i,j;
		int n = 0;
		GeneNode ni;
		for(i = c0.iterator(); i.hasNext();) {
		    ni = (GeneNode)i.next();
		    for(j = c1.iterator(); j!=null && j.hasNext();) {
			if( ni.rootedEquals((GeneNode)j.next()) ) {
			    n++;
			    j.remove();
			    j=null;
			}
		    }
		}
		if(c0.size()!=n) {
		    return false;

		} else {
		    return true;
		}
	    }
	}
    }

    /**
     * Sets the edge value for the edge that belongs to this node.
     *
     * @param edge is the new edge value.
     * @return this for chaining
     */
    public GeneNode setEdgeValue(double edge) {
	this.edge = edge;
	return this;
    }

    /**
     * Sets the edge type and edge value for this node. 
     *
     * @param edgeType is the edge type
     * @param edge is the new edge value
     * @return this for chaining
     */
    public GeneNode setEdge(EdgeType edgeType, double edge) {
	this.setEdgeType(edgeType);
	this.edge = edge;
	return this;
    }

    /**
     * Sets the edge type for this node.
     *
     * @param edgeType is the edge type for this node
     * @return this for chaining
     */    
    public GeneNode setEdgeType(EdgeType edgeType) {
	/*
    	if(edgeType.isRootType()) {
	    throw new FatalException("Can not set a nodes type to root type.");
	}
	*/
	this.edgeType = edgeType;
	return this;
    }

    /**
     * Sets the number of duplications this node causes.
     *
     * @param numberOfDuplications is the new number of duplications this
     *                             node causes given a species tree
     */
    protected void setDuplicationsNode(int numberOfDuplications) {
    	this.duplications = numberOfDuplications;
    }

    /**
     * Returns the number of duplications this node causes given a
     * species tree.
     *
     * @return the number of duplications
     */
    public int getDuplicationsNode() {
	return this.duplications;
    }

    /**
     * Setting the number of gene losses this node causes.
     *
     * @param geneLosses is the number of gene losses this
     *                   node causes given a species tree.
     */    
    protected void setGeneLossesNode(int geneLosses) {
    	this.geneLosses = geneLosses;
    }

    /**
     * Returns the number of gene losses this node causes given a
     * species tree.
     *
     * @return the number of gene losses
     */
    public int getGeneLossesNode() {
	return this.geneLosses;
    }

    /**
     * Returns true if this node is duplicated.
     *
     * @return the same as {@link #getDuplicationsNode}>0
     */
    public boolean duplicatedNode() {
    	return getDuplicationsNode()>0;
    }

    /**
     * See the m-map in the article.
     *
     * @return the mapped species node
     */
    public SpeciesNode getM() {
	if(this.m==null) {
	    this.m = createM();
	}
	return this.m;
    }
    private SpeciesNode createM() {
	if(!hasChildren()) {
	    throw new FatalException("The m for this leaf has not been set.");
	}
	Iterator i = getChildren().iterator();
	SpeciesNode ns = ((GeneNode)i.next()).getM();
	while(i.hasNext()) {
	    ns = SpeciesNode.computeLca(ns, ((GeneNode)i.next()).getM());
	}
	return ns;
    } 

    /**
     * Returns true if M has been set.
     *
     * @return true if M has been set
     */
    public boolean hasM() {
	return this.m!=null;
    }
    
    /**
     * See the M-map in the article.
     *
     * @return a set of species nodes.
     * @see SpeciesNode
     */
    public Set getMM() {
	if(this.mm==null) {
	    this.mm = createMM();
	}
	return this.mm;
    }
    private Set createMM() {
	SpeciesNode m = getM();
	int nc = m.numChildren();
	HashSet mm = new HashSet();
	if(nc>0) {
	    if(nc>2) {
		for(Iterator i = getChildren().iterator();i.hasNext();) {
		    for(Iterator j = ((GeneNode)i.next()).getMM().iterator()
			    ;j.hasNext();) {
			SpeciesNode mc = (SpeciesNode)j.next();
			while(mc.getParent()!=m) {
			    mc = (SpeciesNode)mc.getParent();
			}
			mm.add(mc);
		    }
		}
	    } else {
		mm.addAll(m.getChildren());
	    }
	} else {
	    mm.add(m);
	}
	return mm;
    }

    /**
     * Returns true if the M-map has been set.
     *
     * @return true if the M-map has been set
     */
    public boolean hasMM() {
	return this.mm!=null;
    }

    /**
     * See the SL function in the article.
     *
     * @return a set of species nodes.
     * @see SpeciesNode
     */
    public Set getSL() {
	if(this.sl==null) {
	    this.sl = createSL();
	}
	return this.sl;
    }
    private Set createSL() {
	SpeciesNode m = getM();
	if(m.numChildren()<3) {
	    return m.getLeaves();
	}
	Set sl = new HashSet();
	for(Iterator i = getMM().iterator();i.hasNext();) {
	    SpeciesNode mc = (SpeciesNode)i.next();
	    sl.addAll(mc.getLeaves());
	}
	return sl;
    }
    
    protected void setSL(Set sl) {
	this.sl = sl;
    }

    /**
     * Returns true if SL has been set.
     *
     * @return true if SL has been set
     */
    public boolean hasSL() {
	return this.sl!=null;
    }
    
    /**
     * Returns the child iterator for this node.
     *
     * @return a child iterator.
     * @see ChildIterator
     */
    public ChildIterator getChildIterator() {
	return new ChildIterator(this);
    }

    /**
     * Returns the number of duplications and losses in this subtree.
     *
     * @return the number of duplications and losses in this subtree.
     */
    public Mutation countMutation() {
	return countMutationRec(new Mutation(0, 0));
    }
    private Mutation countMutationRec(Mutation mutate) {
	if(hasChildren()) {
	    mutate.addDuplication(getDuplicationsNode());
	    mutate.addLoss(getGeneLossesNode());
	    for(Iterator i = getChildIterator();i.hasNext();) {
		((GeneNode)i.next()).countMutationRec(mutate);
	    }
	}
	return mutate;
    }
    
    /**
     * Returns the number of losses in this subtree.
     *
     * @return the number of losses in this subtree.
     */
    public int countGeneLosses() {
	if(hasChildren()) {
	    int gene = getGeneLossesNode();
	    for(Iterator i = getChildIterator();i.hasNext();) {
		gene += ((GeneNode)i.next()).countGeneLosses();
	    }
	    return gene;
	}
	return 0;
    }
    
    /**
     * Returns the number of duplications in this subtree.
     *
     * @return the number of duplications in this subtree.
     */
    public int countDuplications() {
	if(hasChildren()) {
	    int dup = getDuplicationsNode();
	    for(Iterator i = getChildIterator();i.hasNext();) {
		dup += ((GeneNode)i.next()).countDuplications();
	    }
	    return dup;
	}
	return 0;
    }

    /**
     * Returns the set of all duplicated nodes in this subtree.
     *
     * @return the set of all duplicated nodes in this subtree
     */
    public Set getDuplicatedNodes() {
	return duplicatedNodesRec(new HashSet());
    }
    private Set duplicatedNodesRec(Set d) {
	if(this.duplicatedNode()) {
	    d.add(this);
	}
	if(hasChildren()) {
	    for(Iterator i=getChildIterator();i.hasNext();) {
		((GeneNode)i.next()).duplicatedNodesRec(d);
	    }
	}
	return d;
    }

    
    /**
     * Sets the collapsed flag to given value.
     *
     * @param collapsedEdge is new value of the collapsed flag
     * @return this gene node for chaining.
     */
    public GeneNode setCollapsedEdge(boolean collapsedEdge) {
    	this.collapsedEdge = collapsedEdge;
    	return this;
    }

    /**
     * Returns true if the node is flagged collapsed.
     *
     * @return the collapsed flag attribute.
     */
    public boolean getCollapsedEdge() {
    	return this.collapsedEdge;
    }

    /**
     * Returns the edge value.
     *
     * @return the edge value
     */
    public double getEdgeValue() {
	if(isRoot()) {
	    return Double.NaN;
	} else {
	    return this.edge;
	}
    }

    /**
     *  Implements setLabel(.) from Node.
     *
     *  @param label is the label number that this internal node will
     *               be set to.
     *  @return An abstract node for chaining.
     *  @throws FatalException when the label is positive.
     *  @see Node
     */
    protected Node setLabel(int label) {
	if(label>=0) {
	    throw new FatalException("Internal node can not be positive.");
	}
	super.setLabel(label);
	return this;
    }

    /**
     * Collapsing all internal nodes with only one child node.
     *
     * Replacing the same method in Node in order to join edges when
     * removing a node with one child.
     *
     * @return the new root.
     */
    public Node pruneOneChildNodes() {
	if(numChildren()>1) {
	    Node n,nn;
	    Collection toAdd = null;
	    Collection toRemove = null;
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		n = (Node)i.next();
		nn = n.pruneOneChildNodes();
		if(!n.equals(nn)) {
		    // n r en enbarnsfrlder
		    if(toAdd==null) {
			toAdd = new Vector();
			toRemove = new Vector();
		    }
		    toRemove.add(n);
		    toAdd.add(nn);
		}
	    }
	    if(toAdd!=null) {
		Iterator add = toAdd.iterator();
		Iterator rem = toRemove.iterator();
		while(add.hasNext()) {
		    removeChild((Node)rem.next());
		    addChild((Node)add.next());
		}
	    }
	    return this;
	} else if(numChildren()==1){
	    // stter vrdet p nn beroende p n och vilken
	    // typ av kant det r.
	    GeneNode nn = (GeneNode)((GeneNode)getChildren().iterator().next())
		.pruneOneChildNodes();
	    if(!this.isRoot()) {
		nn.setEdgeValue(getEdgeType().joinEdges(nn,this));
	    }
	    return nn;
	}
	throw new FatalException();
    }

    /**
     *  See Node
     */
    protected String getEdgeInfo() {
	StringBuffer sb = new StringBuffer();
	//sb.append("(");
	if(getDuplicationsNode()>0) {
	    sb.append("D");
	    sb.append(new Integer(getDuplicationsNode()).toString());
	}
	if(getGeneLossesNode()>0) {
	    sb.append("L");
	    sb.append(new Integer(getGeneLossesNode()).toString());
	}	
	if(!isRoot() && isEdgeWeak()) {
	    sb.append("W");
	}
	if(!isRoot() && getCollapsedEdge()) {
	    sb.append("C");
	}
	
	return sb.toString();
    }
    
    /**
     *  Returns a string representation of the edge. Small letter
     *  mean zero or flag is set to false. For instance 'cD2:BO=97.0'
     *  means that its collapsed flag is false, it has two duplications
     *  and it has a bootstrap value of 97.0.
     *
     *  @return A string representation of the edge.
     */
    protected String toStringEdge() {
	String d,c;
	if(getCollapsedEdge()) {
	    c = "C";
	} else {
	    c = "c";
	}
	if(duplicatedNode()) {
	    d = "D"+getDuplicationsNode();
	} else {
	    d = "d";
	}
    	return c+d+":"+getEdgeType().getShortName() + "=" + getEdgeValue();
    }

    /**
     * Returns true if the edge for this node is weak, else false.
     *
     * @return true if the edge for this node is weak, else false.
     */
    public boolean isEdgeWeak() {
	return getEdgeType().isEdgeWeak(this);
    }
    
    /**
     * Returns the next unique node with edge with give value and
     * edge type.
     *
     * @param value is the edge value
     * @param edgeType is the edge type
     * @return the new gene node.
     */
    public static GeneNode nextNode(EdgeType edgeType, double value) {
	return new GeneNode(edgeType, -1*nextUnique(), value);
    }

    /**
     * Return the next unique node with given edge type.
     *
     *  @return the new gene node.
     */    
    public static GeneNode nextNode(EdgeType edgeType) {
	return nextNode(edgeType, edgeType.getDefaultNodeEdge());
    }
 

    // *****  Save for later implementation of parser *****

    /**
     *  Return the subtree in Screiber-format.
     *
     *  @param label The leaves have two labels. If set to true the
     *               unique label will be printed. If set to false the
     *               id number from will be used.
     *  @return The subtree in Screiber-format.
     */
    /*
    public String toDarwin(boolean label)  {
	StringBuffer sb = new StringBuffer();
	sb.append("[");
	if(hasChildren()) {
	    this.recToDarwin(sb,-1,false,label);
	} else {
	    if(!label && this instanceof GeneLeaf) {
		sb.append( ((GeneLeaf)this).labelFile());
	    } else {
		sb.append( getLabel());
	    }
	}
	sb.append("]");
	return sb.toString();
    }
    private int recToDarwin(StringBuffer sb, int nextLabel,
			    boolean addComma, boolean label) {
	GeneNode n; Iterator iter;
	for(iter=getChildren().iterator();iter.hasNext();) {
	    n = (GeneNode)iter.next();
	    if(!(n instanceof GeneLeaf)) {
		nextLabel = n.recToDarwin(sb,nextLabel,true,label);
	    }
	}
	this.setLabel(nextLabel--);
	
	sb.append("[");
	for(iter=getChildren().iterator();iter.hasNext();) {
	    n = (GeneNode)iter.next();
	    if(n instanceof GeneLeaf) {
		if(!label) {
		    sb.append( ((GeneLeaf)n).labelFile());
		} else {
		    sb.append( ((GeneLeaf)n).getLabel());
		}
	    } else {
		sb.append('[');
		sb.append((-n.getLabel()));
		sb.append(']');
	    }
	    if(iter.hasNext()) {
		sb.append(",");
	    }
	}
	sb.append("]");
	if(addComma) {
	    sb.append(",");
	}
	return nextLabel;
    }
    */
	
}


