
package softparsmap;


import java.util.*;

/**
 * This interface is responsible of handling inparalogous.
 */
public interface Inparalogous {
    
    /**
     * Returns a set of nodes N where all nodes n in N only have
     * in-paralogous sequences in its subtree.
     *
     * @param family is the family of the gene tree
     * @param node is the root node of the gene tree
     * @return a set of internal nodes where all are in-paralogous
     * @see GeneNode
     * @see Family
     */
    public Set getInparalogous(Family family, GeneNode node);

	
    /**
     * Returns true if all gene leaves are coming from the
     * same species. That is if all genes are in-paralogous.
     *
     * @param node is the root of the subtree
     * @return true if all genes in this subtree are
     *         in-paralogous.
     * @see GeneNode
     */
    public boolean allLeavesIsFromSameSpecies(GeneNode node);

    
    /**
     * Removes all inparalogous in the gene tree. 
     *
     * @param family is the family of the gene tree
     * @param root is the root node of the gene tree
     * @return the root node for the tree where all inparalogous
     *         are removed.
     * @see Family
     * @see GeneNode
     */
    public GeneNode removeInparalogous(Family family, GeneNode root);

    
    /**
     * Returns true iff inparalogous are removed after each mid point
     * re-root, before duplications and loss is counted, else false.
     *
     * @return true iff inparalogous are removed while minimizing mutation.
     */
    public boolean removeWhileMinimizingMutation();

    
    /**
     * Returns true iff inparalogous are removed before saving the final
     * gene tree.
     *
     * @return true iff inparalogous are removed before saving the gene tree
     */
    public boolean removeBeforeSaving();

    
    /**
     * When inparalogous are removed one of the gene leaves will be
     * chosen as the final sequence to represent the species. This order
     * that is set here will then be used.
     *
     * @param root is the root of the gene tree
     * @see GeneNode
     */
    public void setInparalogousOrder(GeneNode root);

    
    /**
     * Returns a comparator that compare gene leaves. If used when
     * sorting the most preferred should come first.
     *
     * @param family is the family that this comparator will be used on
     * @return a comparator that will be used to choose the right sequence
     * @see Family
     */
    public Comparator getLeafComparator(Family family);
    
}
