
package softparsmap;


import java.util.*;
import java.io.*;

/**
 * The comparator created below will decide which sequence that
 * will be chosen. For details see method description.
 */
public class InparalogousStandard extends AbstractInparalogous {

    /**
     * Returns a comparator that compare gene leaves. If used when
     * sorting the most preferred comes first.
     *
     * Given two sequences A, B, the preferred will be chosen as follows.
     * <ol>
     *   <li>if A is complete and B is not, A is preferred</li>
     *   <li>if above could not decide then if B is complete and A is not,
     *       B is preferred</li>
     *   <li>if above could not decide then if A is longer then B then A
     *       is preferred</li>
     *   <li>if above could not decide then if B is longer then A then B
     *       is preferred</li>
     *   <li>if above could not decide then if A's GI number is higher then B's
     *       GI number, then A is preferred</li>
     *   <li>if above could not decide then if B's GI number is higher then A's
     *       GI number, then B is preferred</li>
     *   <li>if above could not decide then A is preferred</li>
     * </ol>
     * @param family is the family that this comparator will be used on
     * @return a comparator that will be used to choose the right sequence
     * @see Family
     */
    public Comparator getLeafComparator(Family family) {
	return new Comparator() {
		public Family family = null;
		public int compare(Object o1, Object o2) {
		    DataSource dst = (DataSource)family
			.getFamilyGroup().getDataSource();
		    GeneLeaf l1 = (GeneLeaf)o1;
		    GeneLeaf l2 = (GeneLeaf)o2;
		    SequenceData sd1 = dst.getSequenceData(l1);
		    SequenceData sd2 = dst.getSequenceData(l2);
		    boolean cl1 = sd1.isCompleteSequence();
		    boolean cl2 = sd2.isCompleteSequence();
		    if(cl1 && !cl2) {
			return -1;
		    }
		    if(cl2 && !cl1) {
			return 1;
		    }		    
		    if(!cl1 && !cl2) {
			int pro = sd2.getSequenceLength() -
			    sd1.getSequenceLength();
			if(pro!=0) {
			    return pro;
			}
		    }
		    return sd2.getGiNumber() - sd1.getGiNumber();
		}
		public Comparator setFamily(Family family) {
		    if(family==null) {
			throw new FatalException
			    ("The family can not be null.");
		    }
		    this.family = family;
		    return this;
		}
	    }.setFamily(family);
    }
    
}
