
package softparsmap;


import java.io.*;
import java.util.*;
import java.lang.*;

/**
 * This class contains the code that implements the soft
 * interpretation of gene duplication and loss. When inserting
 * species splits this class will also be used when reducing
 * uncertainties.
 */
public class MuNode extends GeneNode {

    protected static int MAX_NUMBER_OF_PARTITION = 1000;
    
    private GeneNode _linkedNode;

    protected MuNode() {
	setNewLabel();
    }

    protected MuNode(GeneNode node) {
	this();
	_linkedNode = node;
    }

    protected Map getAllInfo() {
	Map info = super.getAllInfo();
	if(_linkedNode!=null) {
	    info.put("L.N.",_linkedNode.toString());
	}
	if(_linkedNode==null) {
	    info.put("E.v."," -");
	}
	return info;
    }
    
    public SpeciesNode getM() {
	if(_linkedNode!=null) {
	    return _linkedNode.getM();
	}
	return super.getM();
    }

    public Set getMM() {
	if(_linkedNode!=null) {
	    return _linkedNode.getMM();
	}
	return super.getMM();
    }

    public Set getSL() {
	if(_linkedNode!=null) {
	    return _linkedNode.getSL();
	}
	return super.getSL();
    }

    public boolean isEdgeWeak() {
	if(_linkedNode!=null) {
	    return _linkedNode.isEdgeWeak();
	}
	return false;
    }
    
    protected boolean hasLinkedNode() {
	return _linkedNode!=null;
    }

    protected GeneNode getLinkedNode() {
	return _linkedNode;
    }
    
    protected void setLinkedNode(GeneNode node) {
	_linkedNode = node;
    }
    
    protected static MuNode nextNode() {
	return new MuNode();
    }

    protected static MuNode nextNode(GeneNode node) {
	return new MuNode(node);
    }

    /*****   Mutation methods   *****/
   
    protected static Mutation computeMutation(Iterator childIterator) {
	//System.out.println("* computeMutation 1");
	List partition = computePartition(childIterator,
					  MAX_NUMBER_OF_PARTITION);
	//System.out.println("* computeMutation 2");
	int loss = computeLossBinary(buildGeneTree(partition));
	//System.out.println("* computeMutation 3");
	return new Mutation(partition.size()-1, loss);
    }

    
    /*****   Loss methods   *****/

    protected static MuNode buildGeneTree(List partition) {
	//System.out.println("* partition="+partition);
	Comparator comp = new Comparator() {
		public int compare(Object o1, Object o2) {
		    return ((MuNode)o1).getSL().size()-
			((MuNode)o2).getSL().size();
		}
	    };
	SortedCollection trees = new SortedCollection(comp);
	for(Iterator i = partition.iterator();i.hasNext();) {
	    List geneNodes = (List)i.next();
	    MuNode node = buildGeneTreeUsingSpeciesTree(geneNodes);
	    //System.out.println("* geneNodes="+geneNodes);
	    //System.out.println("* node\n"+node.toStringAll());
	    trees.add(node);
	}
	while(trees.size()>1) {
	    Iterator i = trees.iterator();
	    MuNode g1 = (MuNode)i.next();
	    boolean foundPair = false;
	    while(!foundPair && i.hasNext()) {
		MuNode g2 = (MuNode)i.next();
		MuNode g = (MuNode)nextNode();
		g.addChild(g1);
		g.addChild(g2);
		if(moveDuplicationsDownSingle(g, g1, g2)>0) {
		    trees.remove(g1);
		    trees.remove(g2);
		    trees.add(g);
		    foundPair = true;
		} else {
		    g.removeChild(g1);
		    g.removeChild(g2);
		}
	    }
	    if(!foundPair) {
		Iterator k = trees.iterator();
		g1 = (MuNode)k.next();
		MuNode g2 = (MuNode)k.next();
		MuNode g = (MuNode)nextNode();
		g.addChild(g1);
		g.addChild(g2);
		trees.remove(g1);
		trees.remove(g2);
		trees.add(g);
	    }
	}
	// moving all duplications as far down as possible
	MuNode root = (MuNode)trees.first();
	//System.out.println("Before moving down\n"+root.toStringAll());
	moveDuplicationsDown(root);
	//System.out.println("After moving down\n"+root.toStringAll());
	return root;
    }

    protected static MuNode buildGeneTreeUsingSpeciesTree(List geneNodes) {
	Map speToGen = new Hashtable();
	// finding and mapping species root
	Iterator iter = geneNodes.iterator();
	SpeciesNode rootM = ((GeneNode)iter.next()).getM();
	while(iter.hasNext()) {
	    rootM = SpeciesNode.computeLca(rootM,
					   ((GeneNode)iter.next()).getM());
	}
	MuNode root = (MuNode)nextNode();
	speToGen.put(rootM, root);
	//System.out.println("*  mu root="+root+", M root="+rootM);
	
	// building gene tree
	for(Iterator i = geneNodes.iterator();i.hasNext();) {
	    //System.out.println("*  speToGen="+speToGen);
	    GeneNode gi = (GeneNode)i.next();
	    MuNode g = (MuNode)MuNode.nextNode(gi);
	    SpeciesNode s = gi.getM();
	    if(speToGen.containsKey(s)) {
		MuNode gp = (MuNode)speToGen.get(s);
		gp.addChild(g);
	    } else {
		while(!speToGen.containsKey(s)) {
		    MuNode gp = (MuNode)MuNode.nextNode();
		    gp.addChild(g);
		    speToGen.put(s, gp);
		    g = gp;
		    s = (SpeciesNode)s.getParent();
		}
		MuNode gp = (MuNode)speToGen.get(s);
		gp.addChild(g);
	    }
	}
	//System.out.println("before making binary \n"+root.toStringAll());
	// make it binary
	return removeSingleChildren(root);
    }

    protected static MuNode removeSingleChildren(MuNode g) {
	//System.out.println("* makebinary in g="+g);
	if(!g.hasChildren()) {
	    return g;
	}
	if(g.numChildren()==1) {
	    MuNode child = (MuNode)g.getChildren().iterator().next();
	    g.removeChild(child);
	    if(g.getParent()!=null) {
		g.getParent().addChild(child);
		g.getParent().removeChild(g);
	    }
	    return removeSingleChildren(child);
	}
	for(Iterator i = new Vector(g.getChildren()).iterator();i.hasNext();) {
	    removeSingleChildren((MuNode)i.next());
	}
	return g;
    }


    
    protected static int moveDuplicationsDown(MuNode g) {
	//System.out.println("* in 1 g="+g);
	if(g.numChildren()!=2) {
	    return 0;
	}
	
	// recursion, bottom first
	MuNode g1, g2;
	Iterator i = g.getChildren().iterator();
	g1 = (MuNode)i.next();
	g2 = (MuNode)i.next();
	int n = moveDuplicationsDown(g1);
	n += moveDuplicationsDown(g2);
	
	// checking so that this is duplicated.
	if(!anyInCommon(g1.getSL(), g2.getSL())) {
	    return n;
	}
	return moveDuplicationsDownSingle(g, g1,g2) + n;
    }
    protected static int moveDuplicationsDownSingle(MuNode g,
						 MuNode g1, MuNode g2) {
	//System.out.println("* in 2 g,g1,g2="+g+","+g1+","+g2);
	if(g1.hasChildren()) {
	    for(Iterator i = g1.getChildren().iterator();i.hasNext();) {
		MuNode g1i = (MuNode)i.next();
		int n = moveDuplicationsDownTriplet(g,g1,g1i,g2);
		if(n>0) {
		    return n;
		}
	    }
	}
	if(g2.hasChildren()) {
	    for(Iterator i = g2.getChildren().iterator();i.hasNext();) {
		MuNode g2i = (MuNode)i.next();
		int n = moveDuplicationsDownTriplet(g,g2,g2i,g1);
		if(n>0) {
		    return n;
		}
	    }
	}
	return 0;
    }  
    protected static int moveDuplicationsDownTriplet(MuNode g,
						  MuNode g1, MuNode g1c,
						  MuNode g2) {
	//System.out.println("* in 3 g="+g+", g1="+g1+", g1c="+g1c+", g2="+g2);

	// we check |g^*1/\g1c| = |(Chi(g1)-g1c+g2)/\g1c| = 0
	Set species = new HashSet();
	for(Iterator i = g1.getChildren().iterator();i.hasNext();) {
	    MuNode g1i = (MuNode)i.next();
	    if(!g1i.equals(g1c)) {
		species.addAll(g1i.getMM());
	    }
	}
	species.addAll(g2.getMM());
	Iterator iter = species.iterator();
	SpeciesNode rootM = (SpeciesNode)iter.next();
	while(iter.hasNext()) {
	    rootM = SpeciesNode.computeLca(rootM, (SpeciesNode)iter.next());
	}
	Set mm = new HashSet();
	if(species.contains(rootM)) {
	    if(rootM.hasChildren()) {
		mm.addAll(rootM.getChildren());
	    } else {
		mm.add(rootM);
	    }
	} else {
	    for(Iterator i = species.iterator();i.hasNext();) {
		SpeciesNode s = (SpeciesNode)i.next();
		while(s.getParent()!=rootM) {
		    s = (SpeciesNode)s.getParent();
		}
		mm.add(s);
	    }
	}
	Set sl1 = new HashSet();
	for(Iterator i = mm.iterator();i.hasNext();) {
	    sl1.addAll(((SpeciesNode)i.next()).getLeaves());
	}
	if(anyInCommon(sl1, g1c.getSL())) {
	    return 0;
	}
	
	//System.out.println("* swiching g2="+g2+" and g1c="+g1c);
	// swiching g2 and g12
	if(g1.numChildren()>2) {
	    MuNode newg11 = nextNode();
	    for(Iterator i = new Vector(g1.getChildren()).iterator()
		    ;i.hasNext();) {
		MuNode g1i = (MuNode)i.next();
		if(!g1i.equals(g1c)) {
		    //System.out.println("* adding gli="+g1i+" to g1 "+g1);
		    g1.removeChild(g1i);
		    newg11.addChild(g1i);
		}
	    }
	    //System.out.println("* newg11="+newg11);
	    g1.addChild(newg11);
	    //throw new FatalException("stop");
	}
	g1.removeChild(g1c);
	g.removeChild(g2);
	g1.addChild(g2);
	g.addChild(g1c);
    
	// reset m, M and SL in g1 and more
	g1.purgeThis();

	//System.out.println("* new tree\n"+g2.root().toStringAll());
	
	// sending it on
	Iterator i = g1.getChildren().iterator();
	return moveDuplicationsDownSingle
	    (g1, (MuNode)i.next(), (MuNode)i.next()) + 1;
    }

    protected static int computeLossBinary(MuNode g) {
	if(!g.hasChildren()) {
	    return 0;
	}
	int loss = 0;
	if(g.numChildren()>2) {
	    for(Iterator i = g.getChildren().iterator();i.hasNext();) {
		loss += computeLossBinary((MuNode)i.next());
	    }
	} else {
	    Iterator i = g.getChildren().iterator();
	    MuNode g1 = (MuNode)i.next();
	    MuNode g2 = (MuNode)i.next();
	    boolean dup = anyInCommon(g1.getSL(), g2.getSL());
	    loss += computeLossBinary(g, g1, dup);
	    loss += computeLossBinary(g, g2, dup);
	    loss += computeLossBinary(g1);
	    loss += computeLossBinary(g2);
	    //System.out.println("* loss("+g+")="+loss);
	}
	return loss;
    }
    
    protected static int computeLossBinary(MuNode g, MuNode gc, boolean gDup) {
	int l1 = 0;
	SpeciesNode climber = gc.getM();
	if(!climber.equals(g.getM())) {
	    climber = (SpeciesNode)climber.getParent();
	    while(!climber.equals(g.getM())) {
		l1++;
		climber = (SpeciesNode)climber.getParent();
	    }
	}
	int l2 = 0;
	if( gc.getM().isBelow(g.getM()) &&
	    !( !gc.getM().hasChildren() ||
	       gc.getMM().containsAll(gc.getM().getChildren()) ) ) {
	    l2 = 1;
	}
	int l3 = 1;
	if(!gDup || (gDup && g.getMM().equals(gc.getMM()))) {
	    l3 = 0;
	}
	//System.out.println("*  g.getMM()="+g.getMM()+
	//", gc.getMM()="+gc.getMM());
	//System.out.println("*  loss("+g+","+gc+")=l1+l2+l3="+
	//  l1+"+"+l2+"+"+l3);
	return l1 + l2 + l3;
    }   

    
    /*****   Duplications methods   *****/
    
    protected static List computePartition(Iterator childIterator,
					int maxPartitionRandom) {
	
	// sort gene nodes such that g_j is before g_i in the
	// list if |SL(g_i)|<=|SL(g_j)|
	Comparator comp = new Comparator() {
		public int compare(Object o1, Object o2) {
		    return
			((GeneNode)o2).getSL().size() -
			((GeneNode)o1).getSL().size();
		}
	    };
	SortedCollection sortedChildNodes = new SortedCollection(comp);
	while(childIterator.hasNext()) {
	    sortedChildNodes.add(childIterator.next());
	}
	
	// stepping through sortedGeneNodes and adding them to buckets
	List buckets = new Vector();
	for(Iterator i = sortedChildNodes.iterator();i.hasNext();) {
	    GeneNode gi = (GeneNode)i.next();
	    addGiToFirstBucket(buckets, gi);
	}

	// first time the sorted method will be used, then the
	// random method
	int numberOfTimes = 0;
	while(numberOfTimes<maxPartitionRandom) {
	    // try to verify the minimum size by common species 
	    if(verifyByCommonSpecies(buckets)) {
		return buckets;
	    }
	    
	    // try to verify the minimum size by common species sets
	    if(verifyByCommonSpeciesSets(buckets)) {
		return buckets;
	    }
	    
	    // creating a new buckets using the random method
	    buckets.clear();
	    List tempGeneNodes = new Vector(sortedChildNodes);
	    while(tempGeneNodes.size()>0) {
		int i = (int)(Math.random()*(double)tempGeneNodes.size());
		GeneNode gi = (GeneNode)tempGeneNodes.remove(i);
		addGiToFirstBucket(buckets, gi);
	    }

	    // see if we should try it again.
	    numberOfTimes++;
	} 

	// failed to count number of duplications
	throw new FatalException("Could not build a partition.");
    }

    
    
    // placing gi in the first buckit where for all gj in the
    // bucket SL(gi) is disjoint with SL(gj)
    protected static void addGiToFirstBucket(List buckets, GeneNode gi) {
	for(int b=0;b<buckets.size() && gi!=null;b++) {
	    List bucket = (List)buckets.get(b);
	    boolean giIsDisjointWithAll = true;
	    for(int j=0;j<bucket.size() && giIsDisjointWithAll;j++) {
		GeneNode gj = (GeneNode)bucket.get(j);
		if(anyInCommon(gi.getSL(), gj.getSL())) {
		    giIsDisjointWithAll = false;
		}
	    }
	    if(giIsDisjointWithAll) {
		// add gi to bucket
		bucket.add(gi);
		gi = null;
	    }
	}
	if(gi!=null) {
	    // gi could not be added to any existing bucket. Adding
	    // a new bucket and adding gi to it.
	    List bucket = new Vector();
	    bucket.add(gi);
	    buckets.add(bucket);
	}
    }
    

    
    // Try to verify that the number of buckets can not be smaller
    // by examine if there is a common species in all buckets.
    protected static boolean verifyByCommonSpecies(List buckets) {
	Set commonSpecies = null;
	for(int b=0;b<buckets.size();b++) {
	    List bucket = (List)buckets.get(b);
	    Set speciesInBucket = new HashSet();
	    for(int j=0;j<bucket.size();j++) {		
		GeneNode gj = (GeneNode)bucket.get(j);
		speciesInBucket.addAll(gj.getSL());
	    }
	    if(commonSpecies==null) {
		commonSpecies = speciesInBucket;
	    } else {
		// Retains only the species in commonSpecies that
		// are contained in current bucket.
		commonSpecies.retainAll(speciesInBucket);
		if(commonSpecies.size()==0) {
		    // if commonSpecies is empty, there are no common species.
		    return false;
		}
	    }
	}
	return true; //commonSpecies.size()>0;
    }


    
    // Try to verify that the number of buckets can not be smaller
    // by finding one species set in every bucket such that any two
    // sets of those are disjoint.
    protected static boolean verifyByCommonSpeciesSets(List buckets) {
	return recVerifyByCommonSpeciesSets(buckets, -1, new Vector());
    }
    protected static boolean recVerifyByCommonSpeciesSets(List buckets,
						       int lastBucketIndex,
						       List geneNodes) {
	if(lastBucketIndex+1>=buckets.size()) {
	    // a list of gene nodes has been found since the end
	    // of the buckets has been reached.
	    return true;
	}
	List currentBucket = (List)buckets.get(lastBucketIndex+1);
	for(int i=0;i<currentBucket.size();i++) {
	    GeneNode gi = (GeneNode)currentBucket.get(i);
	    // checking if gi has common species with all gene
	    // nodes in geneNodes
	    boolean commonSpeciesWithAll = true;
	    for(int j=0;j<geneNodes.size() && commonSpeciesWithAll;j++) {
		GeneNode gj = (GeneNode)geneNodes.get(j);
		if(!anyInCommon(gi.getSL(), gj.getSL())) {
		    commonSpeciesWithAll = false;
		}
	    }
	    // if true gene node will be added and sent on in the recursion
	    if(commonSpeciesWithAll) {
		geneNodes.add(gi);
		if(recVerifyByCommonSpeciesSets
		   (buckets, lastBucketIndex+1, geneNodes)) {
		    return true;
		}
		// did not reach all the way so gi is removed
		geneNodes.remove(geneNodes.size()-1);
	    }
	}
	return false;
    }

    /**
     *  Return true if the cut if the sets 'a' and 'b' is non empty,
     *  else false.
     *
     *  @param a is the first set.
     *  @param b is the second set.
     *  @return True if the set 'a' and 'b' have any objects in common.
     */
    protected static boolean anyInCommon(Set a, Set b) {
	for(Iterator i = a.iterator();i.hasNext();) {
	    if( b.contains(i.next()) ) {
		return true;
	    }
	}
	return false;
    }  
    
}
