
package softparsmap;


import java.io.*;
import java.util.*;
import java.lang.*;

/**
 * Define common behavior for gene and species nodes and trees.
 * When writing 'this subtree' below it refers to the tree where
 * this node is the root.
 */
public abstract class Node implements Comparable { 
    
    /**
     *  This map kan be used to save any relevant information about this node.
     */
    private Map _map;

    /**
     *  Every internal node can have as many children as preferred.
     */
    private Set children;

    /**
     *  Every node except the root will have exactly one parent. The root
     *  is defined by setting this attribute to null.
     */
    private Node parent;

    /**
     *  Every node has a unique label. For any two nodes from the
     *  same tree have to have different labels.
     */
    protected int label;

    /**
     *  The set of gene leaves that exists in this subtree.
     */
    private Set leaves = null;

    /**
     *  The set of species leaves that exists in this subtree.
     */
    //private Set species = null;

    /**
     *  Next unique node will have this label.
     */
    private static int nextUnique = 1;

    /**
     * Constructs a gene node with given label.
     *
     * @param label the label for this node. Two labels in the same
     *              tree can not be the same.
     */
    public Node(int label) {
	this.label = label;
    }

    /**
     * Constructor for extending classes. Does nothing.
     */
    protected Node() {
    }

    /**
     * Returns the splits that exist in this tree. This method can
     * only be used on the root.
     *
     * @return the splits that exist in this tree.
     * @see Split
     */
    public List getSplits() {
	if (!isRoot()) {
	    throw new FatalException
		("This method can only be called on the root.");
	}
	List splits = new Vector();
	getSplitRec(splits);
	return splits;
    }
    private void getSplitRec(List splits) {
	if (!isRoot()) {
	    splits.add(new Split(this));
	}
	if (hasChildren()) {
	    for (Iterator i = getChildren().iterator();i.hasNext();) {
		((Node)i.next()).getSplitRec(splits);
	    }
	}
    }

    /**
     * Returns a map that can be used to save any additional information
     * associated with this node.
     * 
     * @return a node specific map.
     */
    public Map getMap() {
	if (_map==null) {
	    _map = new Hashtable();
	}
	return _map;
    }
    
    public void setMap(Map map) {
	_map = map;
    }


    /**
     * Returns true if this tree is consistent with the given tree.
     * A tree T1 is consistent with a tree T2 iff all splits in
     * T1 is consistent with every split in T2.
     *
     * @return true if this tree is consistent with the given tree.
     * @see Split
     */
    public boolean isConsistentWith(Node tree) {
	if (!isRoot() || !tree.isRoot()) {
	    throw new FatalException
		("This node and the given node must be the root of its tree.");
	}
	// if this tree contains more leaves then the given tree, this tree
	// can not be consistent with the given tree.
	if (getLeaves().size()>tree.getLeaves().size()) {
	    return false;
	}
	//
	List splitsThis = getSplits();
	List splitsTree = tree.getSplits();
	//System.out.println("??? splitsThis="+splitsThis.size());
	//System.out.println("??? splitsTree="+splitsTree.size());
	for (Iterator thisI = splitsThis.iterator();thisI.hasNext();) {
	    Split splitThis = (Split)thisI.next();
	    for (Iterator treeI = splitsTree.iterator();treeI.hasNext();) {
		Split splitTree = (Split)treeI.next();
		//System.out.print("??? this="+splitThis.toStringAll()+
		//		 ", other="+splitTree.toStringAll());
		if (!splitThis.isConsistentWith(splitTree)) {
		    //System.out.println("  .... NOT consistent !!");
		    return false;
		    
		}
		//System.out.println("  .... consistent");
	    }
	    thisI.remove();
	}
	return true;
    }
    
    /**
     * Find the last common ancestor in this subtree for the given
     * set of leaves.
     *
     * @param leaves is a subset of all leaves in this tree
     * @return the last common ancestor
     */
    public Node findLCA(Set leaves) {
	if(hasChildren()) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		Node n = (Node)i.next();
		if(n.getLeaves().containsAll(leaves)) {
		    return n.findLCA(leaves);
		}
	    }
	    return this;
	} else {
	    if(leaves.size()==1 && this.equals(leaves.iterator().next())) {
		return this;
	    }
	    return null;
	}
    }
    
    /**
     * Return true if the subtree where this node is the root is binary,
     * else false.
     *
     * @return true if the subtree is binary, else false.
     */
    public boolean isBinaryTree() {
	if(numChildren()==2) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		if(!((Node)i.next()).isBinaryTree()) {
		    return false;
		}
	    }
	    return true;
	} else {
	    return false;
	}
    }
   
    /**
     * Return all nodes that have more then two children in the subtree
     * where this node is root.
     *
     * @return a set of internal nodes that all have more then two
     *         children. 
     */
    protected Set nonBinaryNodes() {
	return nonBinaryNodesRec(new HashSet());
    }
    private Set nonBinaryNodesRec(Set s) {
	if(hasChildren()) {
	    if(numChildren()>2) {
		s.add(this);
	    }
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		((Node)i.next()).nonBinaryNodesRec(s);
	    }
	}
	return s;
    }
    
    /**
     * Return all internal nodes and leaves in this subtree.
     *
     * @return a set containing all internal nodes and leaves of
     *         this subtree.
     */
    public Set getNodes() {
	Set s = new HashSet();
	s.add(this);
	if(hasChildren()) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		s.addAll(((Node)i.next()).getNodes());
	    }
	}
	return s;
    }
    
    /**
     * Return the leaves of this subtree.
     *
     * @return the leaves of this subtree. 
     */
    public Set getLeaves()  {
	if(this.leaves==null) {
	    this.leaves = new HashSet();
	    if(this.hasChildren()) {
		for(Iterator iter=getChildren().iterator();iter.hasNext();) {
		    for(Iterator k=((Node)iter.next()).getLeaves()
			    .iterator();k.hasNext();) {
			if(!this.leaves.add(k.next())) {
			    throw new FatalException
				("There can not be two leaves with the same "+
				 "label.");
			}
		    }
		}		
	    } else {
		this.leaves.add(this);
		//throw new FatalException("An internal node with no "+
		// "children is not a valid node.");
	    }
	}
	return this.leaves;
    }
   
    /**
     * Remove all internal nodes with only one child. This is done
     * on this subtree and the lowest node is retained.
     *
     * @return the new root for the changed subtree.
     */
    public Node pruneOneChildNodes() {
	if(numChildren()>1) {
	    Node n,nn;
	    Object[] chi = getChildren().toArray();
	    for(int i=0;i<chi.length;i++) {
		n = (Node)chi[i];
		nn = n.pruneOneChildNodes();
		if(!n.equals(nn)) {
		    // n r en enbarnsfrlder
		    this.removeChild(n);
		    this.addChild(nn);
		}
	    }
	    return this;
	} else if(numChildren()==1){
	    return ((Node)getChildren().iterator().next())
		.pruneOneChildNodes();
	}
	throw new FatalException();
    }
       
    /**
     *  Will purge all information in the subtree that are
     *  temporarily stored. Using the method purgeThis() on
     *  every node in the subtree.
     *
     *  @return This node for chaining.
     */
    protected Node purgeDown() {
	purgeThis();
	if(hasChildren()) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		((Node)i.next()).purgeDown();
	    }
    	}
    	return this;
    }

    /**
     *  Purge all information in this and its ancestors. Using the
     *  method purgeThis() on every node in the subtree.
     *
     *  @return This node for chaining.
     */
    protected Node purgeUp() {
	Node node = this;
	do {
	    node.purgeThis();
	    node = node.getParent();
	} while(node!=null);
	return this;
    }

    /**
     *  Purge all information in this node that are
     *  temporarily stored. At present that is to set
     *  the attributes 'leaves' and 'species' to null.
     *
     *  @return This node for chaining.
     */
    protected Node purgeThis() {
    	// attributes that are temporary
    	this.leaves = null;
	//this.species = null;	
    	return this;
    }

    /**
     * Returns the root of this tree.
     *
     * @return the root of this tree
     */
    public Node getRoot() {
	Node root = this;
	while(!root.isRoot()) {
	    root = root.getParent();
	}
	return root;
    }
    
    /**
     * Two nodes are equal iff they have the same label.
     *
     * @return true if their labels are the same, else false.
     */
    public boolean equals(Object o) {
	return this.getLabel() == ((Node)o).getLabel();
    }

    /**
     * Returns the label value. Consistent with {@link #equals}.
     *
     * @return the label value.
     */
    public int hashCode() {
	return getLabel();
    }
    
    /**
     * Returns this label number minus the given nodes label number.
     *
     * @param o is the given node
     * @return this label number minus the given nodes label number
     * @see java.lang.Comparable
     */
    public int compareTo(Object o) {
	return this.getLabel() - ((Node)o).getLabel();
    }

    /**
     * Remove the child from this parent and set the child's
     * parent to null.
     *
     * @param child is the child of this parent that will be removed.
     * @return true if the child was removed, else false.
     */
    public boolean removeChild(Node child) {
	if(removeChildWithoutPurge(child)) {
	    this.purgeUp();
	    return true;
	}
	return false;
    }
    

    /**
     * Remove the child from this parent without purging this node
     * and its ancestors.
     *
     * @param child is the child of this parent that will be removed.
     * @return True if the child was removed, else false.
     */    
    private boolean removeChildWithoutPurge(Node child) {
	if(hasChildren()) {
	    if(getChildren().remove(child)) {
		child.parent = null;
		return true;
	    } else {
		return false;
	    }
	}
	return false;
    }

    
    /**
     * Adds a child under this parent and sets the child's parent
     * as this node.
     *
     * @param child is the new child.
     * @return this node for chaining.
     * @throws FatalException if there already exists a child with the
     *         the same label.
     */
    public Node addChild(Node child) {
	this.purgeUp();
	return addChildWithoutPurge(child);
    }
    
    /**
     *  Add a child under this parent without purging this node and
     *  its parent. 
     *
     *  @param child is the new child.
     *  @return This node for chaining.
     *  @throws FatalException if there already exists a child with the
     *          the same label.
     */
    protected Node addChildWithoutPurge(Node child) {
	if(getChildren().add(child)) {
	    child.setParent(this);
	} else {
	    String se = "There is one or more children with label " +
		child.getLabel() + " under node " + getLabel() +".";
	    throw new FatalException(se);
	}
	return this;
    }

    /**
     *  Adds a set of children using the addChild(.) method.
     *
     *  @param childrenSet is a set of children that will be added
     *                     under this node.
     *  @return This node for chaining. 
     */
    protected Node addChildren(Set childrenSet) {
	for(Iterator i=childrenSet.iterator();i.hasNext();) {
	    addChild((Node)i.next());
	}
	return this;
    }

    /**
     *  Setting the parent. This child will not be removed from any
     *  present child set.
     *
     *  @param parent is the new parent.
     *  @return This for chaining.
     */
    protected Node setParent(Node parent) {
	if(parent==null) {
	    this.parent = null;
	    return this;
	}
	if(this.equals(parent)) {
	    throw new FatalException("Can not be parent to it self.");
	}
	this.parent = parent;
	return this;
    }
    
    /**
     * Returns the parent of this node.
     *
     * @return the parent of this node.
     */
    public Node getParent() {
	return this.parent;
    }
    
    /**
     * Returns true if this node is the root of the tree.
     *
     * @return true if this node is the root.
     */
    public boolean isRoot() {
	return this.parent==null;
    }

    /**
     * Returns the label.
     *
     * @return this nodes label
     */
    public int getLabel() {
        return this.label;
    }

    /**
     *  Setting the label.
     *
     *  @param label is the new label.
     *  @return This node for chaining.
     */
    protected Node setLabel(int label) {
	this.label = label;
	return this;
    }
    
    /**
     * Returns the label number.
     *
     * @return the label number
     */
    public String toString() {
	return new Integer(getLabel()).toString();
    }

    /**
     * Return {@link #toStringTree} and {@link #toStringTable}.
     *
     * @return {@link #toStringTree} and {@link #toStringTable}
     */
    public String toStringAll() {
	return toStringTree() + toStringTable();
    }
    
    /**
     * Returns a easy to read tree of this subtree.
     *
     * @return a easy to read tree of this subtree
     */
    public String toStringTree() {
	return this.toStringTreeRec().toString();
    }
    private TreeTextBlock toStringTreeRec() {
	if(hasChildren()) {
	    List allTtb = new Vector();
	    //System.out.println("** this: " + this);
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		allTtb.add( ((Node)i.next()).toStringTreeRec() );
	    }
	    //System.out.println("** this end, allTtb.size(): " +
	    //  allTtb.size());
	    return new TreeTextBlock(allTtb, this);
	} else {
	    return new TreeTextBlock(this);
	}
    }

    /**
     * Returns a table of all the attribute information
     * that is found in this subtree.
     *
     * @return A table containing all information that exists
     *         in this subtree. If attributes are missing it
     *         is because they are null. 
     */
    public String toStringTable() {
	Comparator comp = new Comparator() {
		public int compare(Object o1, Object o2) {
		    String s1 = (String)((Map)o1).get("ORG tag");
		    if(s1==null) {
			s1 = ((Map)o1).get("Label").toString();
		    }
		    String s2 = (String)((Map)o2).get("ORG tag");
		    if(s2==null) {
			s2 = ((Map)o2).get("Label").toString();
		    }
		    return s1.compareTo(s2);
		    //return ((Comparable)((Map)o1).get("Label")).
		    //	compareTo(((Map)o2).get("Label"));
		}
		public boolean equals(Object obj) {
		    return false;
		}
	    };
	SortedCollection allNodes = new SortedCollection(comp);
	// collecting data
	Set headers = new TreeSet();
	for(Iterator i = getNodes().iterator();i.hasNext();) {
	    Map info = ((Node)i.next()).getAllInfo();
	    headers.addAll(info.keySet());
	    allNodes.add(info);
	}
	// fixing header
	headers.remove("Label");
	List headerList = new Vector();
	headerList.add("Label");
	headerList.addAll(headers);
	// making the table
	List table = new Vector();
	table.add(headerList);
	for(Iterator i = allNodes.iterator();i.hasNext();) {
	    Map info = (Map)i.next();
	    List row = new Vector();
	    for(Iterator k = headerList.iterator();k.hasNext();) {
		Object v = (Object)info.get(k.next());
		if(v==null) {
		    row.add(" -");
		} else {
		    row.add(v.toString());
		}
	    }
	    table.add(row);
	}
	return toStringTableHelp("",table);
    }
    private String toStringTableHelp(String b, List table) {
	StringBuffer sb = new StringBuffer();
	int[] rowWidth = new int[((List)table.get(0)).size()];
	for(Iterator iter = table.iterator();iter.hasNext();) {
	    List row = (List)iter.next();
	    for(int i=0;i<rowWidth.length;i++) {
		int len = ((String)row.get(i)).length();
		if(rowWidth[i]<len) {
		    rowWidth[i] = len;
		}
	    }
	}
	String ssr = startStopRow(rowWidth);
	sb.append(b);
	sb.append(ssr);
	for(int r=0;r<table.size();r++) {
	    List row = (List)table.get(r);
	    sb.append(b);
	    sb.append('|');
	    for(int i=0;i<rowWidth.length;i++) {
		String v = row.get(i).toString();
		sb.append(' ');
		sb.append(v);
		sb.append(space(rowWidth[i]-v.length(),' '));
		sb.append(" |");
	    }
	    sb.append("\n");
	    if(r==0) {
		sb.append(b);
		sb.append(ssr);
	    }
	}
	sb.append(b);
	sb.append(ssr);
	return sb.toString();
    }
    private static String startStopRow(int[] rowWidth) {
	StringBuffer sb = new StringBuffer();
	sb.append('+');
	for(int i=0;i<rowWidth.length;i++) {
	    sb.append(space(rowWidth[i]+2,'-'));
	    sb.append('+');
	}
	sb.append("\n");
	return sb.toString();
    }
    private static String space(int n, char c) {
	StringBuffer sb = new StringBuffer(n);
	for(int i=0;i<n;i++) {
	    sb.append(c);
	}
	return sb.toString();
    }
    
    /**
     *  Return a map mapping the name of the attributes for
     *  this nodes and the attribute values.
     *  
     *  @return A map of the attributes for this node.
     */
    protected Map getAllInfo() {
	Map info = new Hashtable();

	String className = this.getClass().getName();
	info.put("Class",className.substring(className.lastIndexOf('.')+1) );
	// protected int label;
	info.put("Label",new Integer(getLabel()).toString());
	// maps
	if (_map!=null && _map.size()>0) {
	    info.put("Map", _map.toString());
	}
	return info;
    }
	
    /**
     * Returns true if this node has children, else false.
     *
     * @return true if thisnode has children, else false.
     */
    public boolean hasChildren() {
	if(children==null) {
	    return false;
	} else {
	    return children.size()>0;
	}
    }

    /**
     * Returns the number of children this node has.
     *
     * @return the number of children this node has.
     */
    public int numChildren() {
	if(hasChildren()) {
	    return this.children.size();
	}
	return 0;
    }

    /**
     * Returns the child set containing the children. If there are no
     * children, an empty set will be created and returned.
     *
     * @return the set of children under this node. If there are no
     *         children an empty set will be created and returned. 
     */
    public Set getChildren() {
	if(!hasChildren()) {
	    this.children = new ChildSet(2);
	}
	return this.children;
    }

    /**
     * Returns the next unique integer. Iterate through all existing
     * integers from 1 to 2^31 and then start over at 1 again. This stream
     * of unique integers are used to label new nodes.
     *
     * @return the next unique integer
     */
    public static int nextUnique() {
	if(nextUnique<0) {
	    nextUnique = 1;
	}
	return nextUnique++;
    }

    public static void setNextUnique(int newNextUnique) {
	nextUnique = newNextUnique;
    }
    
}
