
package softparsmap;


import java.io.*;
import java.util.*;

/**
 *  Defines a species leaf in a species tree.
 */
public class SpeciesLeaf extends SpeciesNode {

    private Set aSequence;
    
    /**
     * Constructs a species leaf given a data source and a label.
     *
     * @param dataSource is the data source used to create this leaf
     * @param label is the label of this species leaf
     */
    public SpeciesLeaf(DataSource dataSource, int label) {
	super(dataSource, label);
	if(label<=0) {
	    throw new FatalException("GeneLeaf label can not be negative.");
	}
    }

    protected Map getAllInfo() {
	Map info = super.getAllInfo();

	// protected Set aSequence;
	if(this.aSequence!=null) {
	    info.put("Seq", aSequence.toString());
	}

	return info;
    }
    
    /**
     * Returns a species leaf with a unique label and with one sequence.
     *
     * @param dataSource is the data source used to create this leaf
     * @param seq is the label of the sequence that will be attached
     *     after creating the new species leaf
     * @return a new species with the sequence with label 'seq' in it.
     */
    public static SpeciesLeaf nextLeafS(DataSource dataSource, int seq) {
	SpeciesLeaf l = nextLeafS(dataSource);
	l.addSequence(new GeneLeaf(dataSource.getPropertyFile()
			       .getEdgeTypeUnknown(), seq));
	return l;
    }

    /**
     * Returns a species leaf with a unique label.
     *
     * @param dataSource is the data source used to create this leaf
     * @return a new species leaf
     */
    public static SpeciesLeaf nextLeafS(DataSource dataSource) {
	SpeciesLeaf l = new SpeciesLeaf(dataSource, nextUnique());
	return l;
    }
    
    public Node pruneOneChildNodes() {
	return this;
    }

    public String toNewick() {
	StringBuffer sb = new StringBuffer();
	sb.append('(');
	for(Iterator i = getSequences().iterator();i.hasNext();) {
	    sb.append( ((GeneLeaf)i.next()).getLabel() );
	    if(i.hasNext()) {
		sb.append(",");
	    }
	}
	sb.append(')');
	return sb.toString();
    }
        
    /**
     * Leaves can not have children so if calling this method will
     * throw a {@link FatalException}.
     *
     * @throws FatalException if called.
     */
    public Set getChildren() {
	throw new FatalException("Leaves can not have children.");
    }
    
    /**
     * Returns the set of gene sequences that belongs to
     * this species.
     *
     * @return a set with at least one instance of {@link GeneLeaf}.
     * @see GeneLeaf
     */
    public Set getSequences() {
	if(this.aSequence==null) {
	    this.aSequence = new ChildSet();
	}
	return this.aSequence;
    }
    
    /**
     * Adds a given sequence.
     *
     * @param leaf is the sequence that will be added
     * @return this for chaining
     * @see GeneLeaf
     */
    public SpeciesLeaf addSequence(GeneLeaf leaf) {
	if(!getSequences().add(leaf)) {
	    throw new FatalException("Gene sequence "+leaf+" already exists.");
	}
	return this;
    }
    
    /**
     * Add a sequence with the the given edge type and label.
     *
     * @param edgeType is the edge type of the gene leaf added
     * @param leafLabel is the label of the new sequence that is added
     * @return this for chaining.
     * @see GeneLeaf
     * @see EdgeType
     */
    public SpeciesLeaf addSequence(EdgeType edgeType, int leafLabel) {
	return addSequence(new GeneLeaf(edgeType, leafLabel));
    }

}
