
package softparsmap;


import java.io.*;
import java.util.*;
import java.lang.*;

/**
 * Defines a species node in a species tree.
 */
public class SpeciesNode extends Node {

    private int poIn = -1;

    private int poOut = -1;

    private int depth = -1;

    private DataSource dataSource = null;
    
    /**
     * Constructs a species node given a data source and the label.
     *
     * @param dataSource is the data source where the species tree was created.
     * @param label is label of the species node
     */
    public SpeciesNode(DataSource dataSource, int label) {
	super(label);
	this.dataSource = dataSource;
    }

    /**
     * Returns the data source for this species node.
     *
     * @return the data source for this species node
     */
    public DataSource getDataSource() {
	return this.dataSource;
    }

    /**
     * Returns the property file.
     *
     * @return the property file
     */
    public PropertyFile getPropertyFile() {
	return getDataSource().getPropertyFile();
    }
    
    protected Map getAllInfo() {
	Map info = super.getAllInfo();
	// protected int poIn = -1;
	// protected int poOut = -1;
	// protected int depth = -1;
	if(this.poIn>-1) {
	    info.put("Pr. IOD", this.poIn+","+this.poOut+","+this.depth);
	}
	// protected int dupCount = -1;
	// protected int dupRound = -1;

	// species name
	try {
	    info.put("Species name",getDataSource().getSpeciesName(this));
	} catch(Exception e) {
	    info.put("Species name","Exception occurred");
	}

	// tax id
	try {
	    if (getDataSource() instanceof DataSourceXmlNcbiTaxonomy) {
		info.put("Tax Id", ((DataSourceXmlNcbiTaxonomy)getDataSource())
			 .getNcbiTagId(this));
	    }
	} catch(Exception e) {}
	
	return info;
    }
   
    /**
     *  Returns the last common ancestor of the given set of species nodes.
     *  Assumes that {@link #setPreorder} has been run.
     *
     *  @param species is a set of species nodes
     *  @return the last common ancestor of the given set of species
     */
    public static SpeciesNode computeLca(Set species) {
	if(species.size()>1) {
	    Iterator i = species.iterator();
	    SpeciesNode n1 = (SpeciesNode)i.next();
	    while(i.hasNext()) {
		n1 = computeLca((SpeciesNode)i.next(), n1);
		//while(!n2.inSubTreeOf(n1)) {
		//    n1 = (SpeciesNode)n1.getParent();
		//}
	    }
	    return n1;
	} else if(species.size()==1) {
	    return (SpeciesNode)species.iterator().next();
	} else {
	    throw new FatalException("Should not be possible.");
	    //return null;
	}
    }

    /**
     *  Returns the last common ancestor of the two given species nodes. 
     *
     *  @return the last common ancestor of the species na and nb.
     */
    public static SpeciesNode computeLca(SpeciesNode na, SpeciesNode nb) {
	if(nb.depth() < na.depth()) {
	    SpeciesNode s = na;
	    na = nb;
	    nb = s;
	}
	while(!na.isOnSameRootPath(nb)) {
	    na = (SpeciesNode)na.getParent();
	}
	return na;
    }


    /**
     *  Return the preorder in number. If preorder has not been called
     *  this will throw an FatalException.
     *
     *  @return The preorder in number.
     *  @throws FatalException if the method preorder has not been called.
     */
    private int poIn() {
	if(this.poIn<=0) {
	    throw new FatalException("The preorder method needs to be "+
				     "called first.");
	}
	return this.poIn;
    }
    
    /**
     *  Return the preorder out number. If preorder has not been called
     *  this will throw an FatalException.
     *
     *  @return The preorder out number.
     *  @throws FatalException if the method preorder has not been called.
     */    
    private int poOut() {
	if(this.poOut<=0) {
	    throw new FatalException("The preorder method needs to be "+
				     "called first.");
	}
	return this.poOut;
    }

    /**
     *  Return the depth. If preorder has not been called
     *  this will throw an FatalException.
     *
     *  @return The depth of this species node measuring from the root. 
     *  @throws FatalException if the method preorder has not been called.
     */
    private int depth() {
	if(this.depth<=0) {
	    throw new FatalException("The preorder method needs to be "+
				     "called first.");
	}	
	return this.depth;
    }
    
    /**
     *  Check if preorder() has been called. Return true if that is
     *  the case else false.
     *
     *  @return True if the preorder method has been called. 
     */
    private boolean isPreorderSet() {
	return !(this.poOut<=0 || this.poIn<=0);
    }
   
    /**
     * Preorder this tree will give every species node two numbers.
     * The methods {@link #isBelow} and {@link #isOnSameRootPath} are
     * two methods that can be run in constant time after this method
     * has been called.
     *
     * @return the last used number
     */
    public int setPreorder() {
	if(isRoot()) {
	    return preorderRecursive(1,1);
	} else {
	    throw new FatalException
		("The method preorder() can only be called on the root node.");
	}
    }
    private int preorderRecursive(int nextNumber, int depth) {
	this.poIn = nextNumber;
	this.depth = depth;
	nextNumber++;
	if(hasChildren()) {
	    depth++;
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		nextNumber = ((SpeciesNode)i.next()).preorderRecursive
		    (nextNumber, depth);
	    }
	}
	this.poOut = nextNumber;
	return nextNumber;	
    }

    /**
     * Returns true if this node is below given node. Run time is constant.
     *
     * @param s is the given species node
     * @return true if this node is below given node
     */
    public boolean isBelow(SpeciesNode s) {
	return s.poIn < this.poIn && this.poIn < s.poOut;
    }

    /**
     * Returns true if this node and the given node is on the same root path.
     * That is, this method returns true if this node is below the given one,
     * or the given one is below this, or its the same node.
     *
     * @return true if this node and the given node is on the same root path
     */
    public boolean isOnSameRootPath(SpeciesNode s) {
	return isBelow(s) || s.isBelow(this) || this.equals(s);
    }

    /**
     * Returns a list of rooted gene trees. These trees are simple
     * the given unrooted gene tree, mid-point rooted on different edges
     * such that duplications and gene loss is the lowest possible and
     * the same for all trees in the list.
     *
     * It is possible to allow weak edges to be collapsed if it will
     * give lower mutation. Duplication is prioritized before gene loss.
     * Edges will not be collapsed, just flagged collapsed.
     *
     * Inparalogous may also be removed before counting mutation. 
     *
     * @param geneTree is the unrooted gene tree
     * @param collapseEdges if true weak edges will be collapsed if it
     *                      gives lower mutation
     * @param inparalogous is the inparalogous handler
     * @return a list of gene trees with the lowest possible number
     *         of duplications and gene loss.
     */    
    public List minimizeMutation(GeneNode geneTree, boolean collapseEdges,
				 Inparalogous inparalogous) {
	List minTreeCol  = new Vector();
	Set nodes = geneTree.getNodes();
	//	System.out.println("**********************");
	//	System.out.println(nodes);
	//	System.out.println(geneTree);
	// tar bort redundanta noder
	if(geneTree.hasChildren()) {
	    for(Iterator i = geneTree.getChildren().iterator();i.hasNext();) {
		nodes.remove(i.next());
	    }
	}

	// initiating paralogous
	boolean removeInparalogous =
	    inparalogous.removeWhileMinimizingMutation();
	if(removeInparalogous) {
	    // setting m
	    this.duplications(geneTree, false, null);
	    // setting gene leaf prio
	    inparalogous.setInparalogousOrder(geneTree);
	}
	
	//	System.out.println(nodes);
	//	System.out.println(geneTree.toStringAll());
	//	System.out.println("**********************");
	// all trees in minTreeCol will have the same # duplications
	// and gene loss and no other tree will have less duplications
	// and gene loss.
	//System.out.println("****************************** " +nodes.size());
	int minDup = Integer.MAX_VALUE;
	int minGene = Integer.MAX_VALUE;
	for(Iterator i=nodes.iterator();i.hasNext();) {
	    GeneNode n = (GeneNode)i.next();
	    // cloning and rerooting tree
	    GeneNode newTree = geneTree.cloneTree().findNode(n).reroot();
	    // if true, removing inparalogous
	    if(removeInparalogous) {
		// removing inparalogous
		newTree = inparalogous
		    .removeInparalogous(newTree.getFamily(),newTree);
	    }
	    // calculationg duplications and loss
	    this.duplications(newTree,collapseEdges,inparalogous);
	    int newDup = newTree.countDuplications();
	    int newGene = newTree.countGeneLosses();
	    //System.out.println("** \n" +newTree.toStringTree());
	    //System.out.println("** geneLoss: " + newGene);
	    if(minDup==newDup && minGene==newGene) {
		minTreeCol.add(newTree);
	    } else if(minDup>newDup || (minDup==newDup && minGene>newGene) ) {
		minTreeCol.clear();
		minTreeCol.add(newTree);
		minDup = newDup;
		minGene = newGene;
	    }
	}
	return minTreeCol;
    }
    // reset flags before calculating the mutations.
    private void resetDuplicationFlags() {
	this.poIn = -1;
	this.poOut = -1;
	//this.dupCount = 0;
	//this.dupRound = 0;
	if(hasChildren()) {
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		((SpeciesNode)i.next()).resetDuplicationFlags();
	    }
	}
    }


    /**
     *  Infers duplications and loss using this species subtree.
     *  Duplications are prioritized before gene loss and the calculation
     *  of gene loss is approximative. The method will collapse weak
     *  edges if it will give a better result and in-paralogous can be
     *  removed before counting mutation. This method can only be called
     *  using the root of the species tree and the gene tree.
     *
     *  @param geneNode is the rooted gene tree
     *  @param collapseWeakEdges if true the method will collapse weak edges
     *                       when it gives lower mutation
     *  @param inparalogous is the inparalogous handler and decides if
     *             inparalogous should be removed before counting mutation
     *  @return the total number of duplications and loss
     */    
    public Mutation inferMutation(GeneNode geneNode, boolean collapseWeakEdges,
				  Inparalogous inparalogous) {
	//
	if(!this.isRoot() || !geneNode.isRoot()) {
	    throw new FatalException("This method can only be called using "+
				     "the root of the species tree and "+
				     "the gene tree. this.isRoot()="+
				     this.isRoot()+", geneNode.isRoot()="+
				     geneNode.isRoot());
	}
	//
	this.duplications(geneNode, collapseWeakEdges, inparalogous);
	//
	Mutation mut = new Mutation(geneNode.countDuplications(),
			  geneNode.countGeneLosses());
	return mut;
    }
    
    /**
     *  See inferMutation(...)
     */    
    private int duplications(GeneNode geneNode, boolean collapseEdges,
			    Inparalogous inparalogous) {
	// setting the M for all leaves in the gene tree
	setMForGeneLeaves(geneNode);
	// preorder all nodes in this species tree
	if(!isPreorderSet()) {
	    ((SpeciesNode)this.getRoot()).setPreorder();
	}
	// resets all flags
	geneNode.resetFlagsDown();
	// if true, setting the collpased flag for all weak edges
	// except those edges that are inparalogous and if dup loss
	// are not to be counted for in the inparalogous.
	if(collapseEdges) {
	    List allWeakEdges = new Vector(geneNode.getInternalNodes());
	    for(Iterator i = allWeakEdges.iterator();i.hasNext();) {
		GeneNode n = (GeneNode)i.next();
		if( n.isRoot() || !n.isEdgeWeak() ) {
		    //if( n.isRoot() || !n.isEdgeWeak() ||
		    //doNotCountMutation(n, inparalogous)) {
		    i.remove();
		} else {
		    n.setCollapsedEdge(true);
		}
	    }
	    // get best mutation
	    duplicationsSub(geneNode, inparalogous);
	    Mutation bestMutation = geneNode.countMutation();
	    //System.out.println("******************************************");
	    // try to restore edges
	    for(int i = allWeakEdges.size()-1;i>=0;i--) {
		GeneNode n = (GeneNode)allWeakEdges.get(i);
		n.setCollapsedEdge(false);
		//System.out.println(geneNode.toStringAll());
		duplicationsSub(geneNode, inparalogous);
		Mutation newMutate = geneNode.countMutation();
		//System.out.println("**** n: "+n+", newMut: "+newMutate);
		if(newMutate.isWorse(bestMutation)) {
		    n.setCollapsedEdge(true);
		    n.setDuplicationsNode(0);
		    n.setGeneLossesNode(0);
		}
	    }
	}	    
	    
	
	// calculating duplications and gene loss
	duplicationsSub(geneNode, inparalogous); //, collapseEdges);
	
	// return the number of duplications.
	return geneNode.countDuplications();
    }

    /*
    private boolean doNotCountMutation(GeneNode geneNode,
				       Inparalogous inparalogous) {
	return inparalogous!=null && !inparalogous.countMutation(geneNode);
    }
    */
    
    private void duplicationsSub(GeneNode geneNode,
				 Inparalogous inparalogous) {
	// if inparalogous is not null and all genes in the subtree of
	// geneNode are in-paralogous, duplications and loss will be
	// ignored for this node and all subnodes.
	//if(doNotCountMutation(geneNode, inparalogous)) {
	//    return;
	//}

	// assume that if 'searchWeakEdges' is true all nodes where the edges
	// is weak the collapsed edge flag is set collapsed. 
	Mutation bestMutate = duplicationsSubCalc(geneNode);
	// return if no children
	if(bestMutate==null) {
	    return;
	}

	// setting this node with duplication and loss
	geneNode.setDuplicationsNode(bestMutate.getDuplication());
	geneNode.setGeneLossesNode(bestMutate.getLoss());
	
	// sending the task on to the sub nodes.
	for(Iterator i = geneNode.getChildIterator();i.hasNext();) {
	    duplicationsSub((GeneNode)i.next(), inparalogous);
	}

    }
    
    private Mutation duplicationsSubCalc(GeneNode geneNode) {
	// return if no shildren
	if(!geneNode.hasChildren()) {
	    return null;
	}
	
	//
	return MuNode.computeMutation(geneNode.getChildIterator());
    }

    private void setMForGeneLeaves(GeneNode geneTree) {
	// build a map from sequences to species
	Map geneToSpe = new Hashtable();
	for(Iterator i = getLeaves().iterator();i.hasNext();) {
	    SpeciesLeaf spe = (SpeciesLeaf)i.next();
	    for(Iterator k=spe.getSequences().iterator();k.hasNext();) {
		geneToSpe.put(k.next(),spe);
	    }
	}
	// setting the 
	for(Iterator i = geneTree.getLeaves().iterator();i.hasNext();) {
	    GeneLeaf l = (GeneLeaf)i.next();
	    SpeciesLeaf ls = (SpeciesLeaf)geneToSpe.get(l);
	    //   System.out.println("***** ls: "+ls+", geneNode: "+geneNode);
	    if(ls==null) {
		throw new FatalException("Found a sequence that did not "+
					 "belong to a species.");
	    }
	    l.setM(ls);
	}
    }
       
    /**
     * Returns the subtree in Newick format.
     *
     * @return the subtree in Newick format.
     */
    public String toNewick() {
	if(hasChildren()) {
	    StringBuffer sb = new StringBuffer();
	    sb.append("(");
	    for(Iterator i=getChildren().iterator();i.hasNext();) {
		sb.append(((SpeciesNode)i.next()).toNewick());
		if(i.hasNext()) {
		    sb.append(",");
		}
	    }
	    sb.append(")");
	    return sb.toString();
	} else {
	    return null;
	}
    }
   
    /**
     * Parse a string in Neiwck format with gene sequences in the
     * inner bracket. Run program {@link InferMutation} to see
     * an example.
     *
     * @param ds is the data source
     * @param treeText is the species tree in text format
     * @return the root for the new species tree
     */
    public static SpeciesNode parseNewick(DataSource ds, String treeText) {
	// remove enters
	treeText = treeText.replaceAll("\n","");
	// remove all spaces
	treeText = treeText.replaceAll(" ","");
	// remove all before first bracket
	treeText = treeText.substring(treeText.indexOf('(')+1,
				      treeText.lastIndexOf(')') );
	// send it on
	return recParseNewick(ds, treeText);
    }
    private static SpeciesNode recParseNewick(DataSource ds, String treeText) {
	//System.out.println("treeText: "+treeText);
	
	// if true there are only sequences left
	if(treeText.indexOf('(')<0) {
	    if(treeText.indexOf(')')>0) {
		throw new FatalException("Following subpart has non matching "+
					 "brackets, text: "+treeText);
	    }
	    // creating species leaf and add the sequences
	    SpeciesLeaf sl = SpeciesLeaf.nextLeafS(ds);
	    StringTokenizer st = new StringTokenizer(treeText, ",");
	    while(st.hasMoreTokens()) {
		sl.addSequence(ds.getPropertyFile().getEdgeTypeUnknown(),
			       Integer.parseInt(st.nextToken()));
	    }
	    // return node
	    return sl;
	}

	// debug help
	if(treeText.charAt(0)!='(') {
	    throw new FatalException();
	}
	
	// creating parent
	SpeciesNode parent = SpeciesNode.nextNodeS(ds);
	
	// sending on subpart
	int brackets = 0;
	int start = 0;
	int p = 0;
	while(p<treeText.length() && p>=0) {
	    // counting brackets
	    if(treeText.charAt(p)=='(') {
		brackets++;
	    } else if(treeText.charAt(p)==')') {
		brackets--;
	    }
	    // if true, creating and adding child.
	    if(brackets==0) {
		parent.addChild
		    (recParseNewick(ds,treeText.substring(start+1, p)));
		// setting parameters for next
		start = treeText.indexOf('(',p);
		p = start;
	    } else {
		// increate p
		p++;
	    }
	}

	// return new node
	return parent;
    }

    /**
     *  Return the next unique labeled species node.
     *
     *  @return The next unique labeled species node.
     */
    public static SpeciesNode nextNodeS(DataSource ds) {
	return new SpeciesNode(ds, -1*nextUnique());
    }
    

}
