
package softparsmap;


import java.util.*;

/**
 * For a tree T=(V,E), let d=(L1,L2) be a split iff L1, L2 [subset_of] L(T),
 * |L1 [cut] L2|=0, L1 [union] L2 = L(T), and there exists an edge e [in] E
 * so if removed we get two tree graphs T1 and T2 where L1 [subset_of] L(T1)
 * and L2 [subset_of].
 */
public class Split {

    private Set _all = null;
    
    private Set _l1 = null;

    private Set _l2 = null;

    /**
     * Constructing a split that exists between the node and its parent.
     * This constructor can not be used on a root.
     *
     * @param node is a node.
     */
    public Split(Node node) {
	if (node.isRoot()) {
	    throw new FatalException
		("This constructor can not be used on a root.");
	} else {
	    _l1 = node.getLeaves();
	    _all = node.getRoot().getLeaves();
	}
    }

    /**
     * Constructor for two sets l1 and l2 defined above.
     *
     * @param l1 is defined above
     * @param l2 is defined above
     */
    public Split(Set l1, Set l2) {
	_l1 = l1;
	_l2 = l2;
    }

    /**
     * Returns the first part of this split.
     *
     * @return the first part of this split
     */
    public Set getL1() {
	return _l1;
    }

    /**
     * Returns the second part of this split.
     *
     * @return the second part of this split
     */
    public Set getL2() {
	if (_l2==null) {
	    _l2 = new HashSet(_all);
	    _l2.removeAll(_l1);
	}
	return _l2;
    }

    /**
     * Returns true if this split is consistent with the given split.
     * A split d=(L1,L2) is consistent with a split d'=(L1',L2') iff
     * L1 [subset_of] L1' or L1 [subset_of] L2' or
     * L2 [subset_of] L1' or L2 [subset_of] L2'
     *
     * @param d is the split tested against this split
     * @return true if this split is consistent with the given split.
     */
    public boolean isConsistentWith(Split d) {
	return
	    d.getL1().containsAll(getL1()) || d.getL2().containsAll(getL1()) ||
	    d.getL1().containsAll(getL2()) || d.getL2().containsAll(getL2());
    }

    public String toStringAll() {
	return getL1().toString()+"|"+getL2().toString();
    }

    /*
    public String toString() {
	return toStringAll();
    }
    */
    
}
