
package softparsmap;


import java.util.*;

/**
 * This class can extract string information from a string template
 * or create a string given a template and data. To clarify see
 * following example. <p>
 * 
 * <pre>
 *   StringTemplate st = new StringTemplate("family{number}.tree", "{", "}");
 *   st.setMarker("{number}", "100");
 *   System.out.println(st.getPaintedTemplate());
 *   st.paintTemplate("family200.tree");
 *   System.out.println(st.getFirstValue("{number}"));
 * </pre>
 * <p>
 * This will print <p>
 * <pre>
 *   family100.tree
 *   200
 * </pre>
 */
public class StringTemplate {

    private String _template = null;
    private String _startMarker = null;
    private String _endMarker = null;

    private List _markers = null;
    private List _delimiters = null;

    /**
     * Construct a string template from a template and what markers
     * start and stop with.
     *
     * @param template is the template
     * @param startMarker defines how markers begin
     * @param stopMarker defines how markers end
     */
    public StringTemplate(String template, String startMarker,
			  String stopMarker) {
	// checking start and stop marker
	if(startMarker.equals(stopMarker) ||
	   startMarker.indexOf(stopMarker)>0 ||
	   stopMarker.indexOf(startMarker)>0) {
	    throw new FatalException
		("Stop and start marker can not be the same and one can "+
		 "not exist in the other.");
	}
	
	// saving data
	_template = template;
	_startMarker = startMarker;
	_endMarker = stopMarker;

	// creating markers and delimiters
	_markers = new Vector();
	  //_mapMarkers = new Hashtable();
	_delimiters = new Vector();
	// pointers
	// ...{...}...{...}...{...}...
	//         |  |    |  |
	//         p0 p1   p2 p3
	if(_template.length()>0) {
	    if(_template.indexOf(startMarker)<0) {
		// no marker
		_delimiters.add(_template);
	    } else {
		// at least one marker
		boolean conti = true;
		int p0=0, p1=0, p2=0, p3=0;
		p1 = _template.indexOf(startMarker,p0);
		while(conti) {
		    // set last pointers
		    p2 = template.indexOf(stopMarker, p1)+stopMarker.length();
		    p3 = template.indexOf(startMarker, p2);
		    // extracting marker
		    Marker m = new Marker(template.substring(p1,p2));
		    //_mapMarkers.put(m.marker, m);
		    _markers.add(m);
		    // extract first delimiter
		    _delimiters.add(template.substring(p0, p1).trim());
		    // check if at end
		    if(p3<0) {
			p3 = template.length();
			conti = false;
			// add last delimiter
			_delimiters.add(template.substring(p2,p3).trim());
		    } else {
			// move pointers on
			p0 = p2;
			p1 = p3;
		    }
		}
	    }
	}
	// checking delimiters
	for(int i=0;i<_delimiters.size();i++) {
	    // only first and last delimiter can be ""
	    if(i>0 && i<_delimiters.size()-1) {
		if(((String)_delimiters.get(i)).length()==0) {
		    throw new FatalException
			("Only first and last delimiter can be empty.");
		}
	    }
	}
	
	// checking markers ??
	
	//System.out.println("*** _markers="+_markers);
	//System.out.println("*** _delimiters="+_delimiters);
	
	
    }

    /**
     * Sets a marker to a new value.
     *
     * @param marker is the marker
     * @param newValue is the new value
     */
    public void setMarker(String marker, String newValue) {
	for(Iterator i = _markers.iterator();i.hasNext();) {
	    Marker m = (Marker)i.next();
	    if(m.marker.equals(marker)) {
		m.value = newValue;
		return;
	    }
	}
	throw new FatalException("Marker '"+marker+"' does not exists in "+
				 "template '"+_template+"'.");
    }

    /**
     * Sets the markers with their own names.
     */
    public void paintTemplateWithMarkerNames() {
	for(Iterator i = _markers.iterator();i.hasNext();) {
	    Marker m = (Marker)i.next();
	    m.value = m.marker;
	}
    }

    /**
     * After markers have been set, calling this method will return
     * the template with its markers painted accordingly.
     *
     * @return the painted marker
     */
    public String getPaintedTemplate() {
	StringBuffer sb = new StringBuffer();
	Marker m = null;
	int i = 0;
	int p0 = 0;
	int p1 = 0;
	while(i<_markers.size()) {
	    m = (Marker)_markers.get(i);
	    p1 = _template.indexOf(m.marker, p0);
	    sb.append(_template.substring(p0,p1));
	    sb.append(m.value);
	    p0 = p1 + m.marker.length();
	    i++;
	}
	sb.append(_template.substring(p0));
	return sb.toString();
    }

    /**
     * If the value matches the template the markers will be set
     * to corresponding values.
     *
     * @param value is string matching the template
     */
    public void paintTemplate(String value) {
	//System.out.println("#### template='"+_template+
	// "', value='"+value+"'");
	if(_template.length()==0) {
	    return;
	}
	//
	int delSize = _delimiters.size();
	int i = 0;
	String delim = (String)_delimiters.get(i);
	//System.out.println("#### 1 delim='"+delim+"'");
	int p0 = 0;
	if(delim.length()>0) {
	    p0 = value.indexOf(delim);
	    //System.out.println("#### 2 p0='"+p0+"'");
	    if (p0<0) {
		throw new FatalException
		    ("Value '"+value+"' does not fit the template '"+
		     _template+"'.");	
	    }
	    p0 = p0+delim.length();
	}
	int p1 = p0;
	i++;
	while(i<delSize) {
	    // extract value
	    delim = (String)_delimiters.get(i);
	    //System.out.println("#### delim='"+delim+"'");
	    if(delim.length()==0) {
		p1 = value.length();
	    } else {
		p1 = value.indexOf(delim, p0);
	    }
	    if(p0<0 || p1<0) {
		throw new FatalException
		    ("Value '"+value+"' does not fit the template '"+
		     _template+"'.");
	    }
	    //System.out.println("####   (p0, p1)=("+p0+", "+p1+")");
	    String mValue = value.substring(p0, p1).trim();
	    ((Marker)_markers.get(i-1)).value = mValue;
	    //System.out.println("####   value='"+mValue+"'");
	    
	    // move pointer and move to next delimiter
	    p0 = p1+delim.length();
	    i++;
	}
    }

    /**
     * Returns true if the marker exists in the template.
     *
     * @return true if the marker exists in the template
     */
    public boolean hasMarker(String marker) {
	for(Iterator i = _markers.iterator();i.hasNext();) {
	    Marker m = (Marker)i.next();
	    if(m.marker.equals(marker)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Returns the value of the first marker named as given.
     *
     * @param marker is the marker
     * @return the value of the marker
     */
    public String getFirstValue(String marker) {
	for(Iterator i = _markers.iterator();i.hasNext();) {
	    Marker m = (Marker)i.next();
	    if(m.marker.equals(marker)) {
		return m.value;
	    }
	}
	throw new FatalException
	    ("Can not find marker='"+marker+"' in template='"+_template+"'.");
    }
    
    private class Marker {
	public String marker = null;
	public String value = null;
	public Marker(String marker) {
	    this.marker = marker;
	}
	public String toString() {
	    return marker+"='"+value+"'";
	}
    }

}

