
package softparsmap;


import ooc.*;
import java.util.*;

/**
 * This parser will parse the Newick format. It is flexible within limits,
 * See the <a href="{@docRoot}/softparsmap/def.xml">def.xml</a>
 * file for more information.
 */
public class TreeParserNewick extends AbstractTreeParser {

    private StringTemplate _dataTemplate;
    private StringTemplate _leafTemplate;
    private String _left;
    private String _right;
    private String _childSeparator;
    private boolean _hasNodeDataTemplateEdgeValue;
    private boolean _hasLeafTemplateEdgeValuereturn;
    private EdgeType _edgeType = null;
    private String _beforeBody = null;
    private String _afterBody = null;

    public void setupInstance() {
	super.setupInstance();
	_left = getAttributeString("recursive_begin");
	_right = getAttributeString("recursive_end");
	_childSeparator = getAttributeString("recursive_child_separator");
	_dataTemplate = 
	    new StringTemplate(getAttributeString("template_node_data"),
					   getBeginMarker(), getEndMarker());
	_leafTemplate = new StringTemplate(getAttributeString("template_leaf"),
					   getBeginMarker(), getEndMarker());
	_hasNodeDataTemplateEdgeValue = 
	    getNodeDataTemplate().hasMarker(getEdgeValueMarker());
	_hasLeafTemplateEdgeValuereturn =
	    getLeafTemplate().hasMarker(getEdgeValueMarker());
	_edgeType = (EdgeType)getAttribute("edge_type");
	_beforeBody = getAttributeString("before_body");
	_afterBody = getAttributeString("after_body");
    }
    

    /**
     * Returns the edge type that exists in the tree. It is defined by
     * the attribute 'edge_type' in the tag.
     *
     * @return an edge type
     * @see EdgeType
     */
    public EdgeType getEdgeType() {
	return _edgeType;
    }

    public void setEdgeType(EdgeType edgeType) {
	_edgeType = edgeType;
    }

    private String getLeft() {
	return _left;
    }

    private String getRight() {
	return _right;
    }

    private String getChildSeparator() {
	return _childSeparator;
    }
      
    private boolean hasNodeDataTemplateEdgeValue() {
	return _hasNodeDataTemplateEdgeValue;
    }
    
    private boolean hasLeafTemplateEdgeValue() {
	return _hasLeafTemplateEdgeValuereturn;
    }
    
    private StringTemplate getNodeDataTemplate() {
	return _dataTemplate;
    }
    
    private StringTemplate getLeafTemplate() {
	return _leafTemplate;
    }

    private String getBeforeBody() {
    	return _beforeBody;
    }
    
    private String getAfterBody() {
    	return _afterBody;
    }

    /** Do not need to speed these up **/
    
    private String getEdgeValueMarker() {
	return getAttributeString("marker_edge_value");
    }

    private String getLeafLabelMarker() {
	return getAttributeString("marker_leaf_label");
    }

    private String getBeginMarker() {
	return getAttributeString("markers_begin");
    }

    private String getEndMarker() {
	return getAttributeString("markers_end");
    }

    public Node parseIn(String tree) {
	// remove new rows
	tree = tree.replaceAll("\n","");

	// root
	GeneNode root = GeneNode.nextNode(getEdgeType());
		
	// remove all before first and last bracket and parse tree
	int p0 = tree.indexOf(getLeft());
	int p1 = tree.lastIndexOf(getRight());
	if(!(0<=p0 && p0<p1)) {
	    throw new FatalException("Could not parse tree.");
	}
	parseInRec(getEdgeType(), root, tree.substring(p0+getLeft().length(),p1));
	// if it was only one leaf then removing above gene node
	if(root.numChildren()==1) {
	    GeneNode child = (GeneNode)root.getChildren().iterator().next();
	    root.removeChild(child);
	    return child;
	}
	return root;
    }
    private void parseInRec(EdgeType edgeType, GeneNode parent,
			    String children) {
	//System.out.println("*** children='"+children+"'");
	// checking if it is a leaf
	if(children.indexOf(getLeft())!=0) {
	    // its a leaf
	    int ends = children.indexOf(getChildSeparator());
	    String leafData = null;
	    if(ends<0) {
		leafData = children.trim();
	    } else {
		leafData = children.substring(0, ends).trim();
	    }
	    // create and add leaf
	    //System.out.println("* leafData='"+leafData+"'");
	    getLeafTemplate().paintTemplate(leafData);
	    int label = Integer.parseInt
		(getLeafTemplate().getFirstValue(getLeafLabelMarker()));
	    double edge = -1;
	    if(hasLeafTemplateEdgeValue()) {
		edge = Double.parseDouble
		    (getLeafTemplate().getFirstValue(getEdgeValueMarker()));
	    } else {
		edge = edgeType.getDefaultLeafEdge();
	    }
	    GeneLeaf leaf = new GeneLeaf(edgeType, label, edge);
	    parent.addChild(leaf);
	    // send it on
	    //System.out.println("** take next child");
	    if(ends>0) {
		parseInRec(edgeType, parent, children.substring
			   (ends+getChildSeparator().length()).trim());
	    }
	    return;
	}

	
	// its a internal node
	int p0 = 0;
	int p1 = 0;
	int leftRight = 0;
	do {
	    int pl = children.indexOf(getLeft(), p1);
	    if(pl<0) {
		pl = Integer.MAX_VALUE;
	    }
	    int pr = children.indexOf(getRight(), p1);
	    if(pr<0) {
		pr = Integer.MAX_VALUE;
	    }
	    //System.out.println("* (pl,pr)=("+pl+","+pr+")");
	    
	    if(pl<pr) {
		leftRight++;
		p1 = pl+getLeft().length();
	    } else {
		leftRight--;
		p1 = pr+getRight().length();
	    }
	    if(leftRight<0) {
		throw new FatalException("Could not parse tree.");
	    }
	} while(leftRight!=0);
	
	// get edge value
	int ends = children.indexOf(getChildSeparator(), p1);
	String nodeData = null;
	if(ends<0) {
	    nodeData = children.substring(p1).trim();
	} else {
	    nodeData = children.substring(p1, ends).trim();
	}
	// create internal node
	getNodeDataTemplate().paintTemplate(nodeData);
	//System.out.println("* nodeData='"+nodeData+"', template='"+
	//   getAttributeString("node_data_template")+"'");
	double edge = -1;
	if(hasNodeDataTemplateEdgeValue()) {
	    edge = Double.parseDouble
		(getNodeDataTemplate().getFirstValue(getEdgeValueMarker()));
	} else {
	    edge = edgeType.getDefaultNodeEdge();
	}
	GeneNode node = GeneNode.nextNode(edgeType, edge);
	parent.addChild(node);
	// if any more children then sending it on
	if(ends>=0) {
	    parseInRec(edgeType, parent, children.substring
		       (ends+getChildSeparator().length()).trim());
	}
	// sending it on down
	String newChildren = children.substring(p0+getLeft().length(),
						p1-getRight().length()).trim();
	parseInRec(edgeType, node, newChildren);
	//
	return;
    }

    
    public String parseOut(Node tree) {
	// paint templates
	getNodeDataTemplate().paintTemplateWithMarkerNames();
	getLeafTemplate().paintTemplateWithMarkerNames();

	// create tree, add before and after and return it
	String stringTree = null;
	if(tree.hasChildren()) {
	    stringTree = parseOutRec((GeneNode)tree);
	} else {
	    stringTree = "("+parseOutRec((GeneNode)tree)+")";
	}
	return getBeforeBody()+stringTree+getAfterBody();
    }
    private String parseOutRec(GeneNode node) {
	StringBuffer sb = new StringBuffer();
	// check if leaf
	if(!node.hasChildren()) {
	    if(hasLeafTemplateEdgeValue()) {
		getLeafTemplate().setMarker
		    (getEdgeValueMarker(),
		     Double.toString(node.getEdgeValue()));
	    }
	    getLeafTemplate().setMarker(getLeafLabelMarker(),
					Integer.toString(node.getLabel()));
	    sb.append(getLeafTemplate().getPaintedTemplate());
	    //
	    return sb.toString();
	}

	// is an intarnal node
	sb.append(getLeft());
	for(Iterator i = node.getChildren().iterator();i.hasNext();) {
	    sb.append(parseOutRec((GeneNode)i.next()));
	    if(i.hasNext()) {
		sb.append(getChildSeparator());
	    }
	}
	sb.append(getRight());
	// data only added if it is not a root
	if(!node.isRoot()) {
	    if(hasNodeDataTemplateEdgeValue()) {
		getNodeDataTemplate().setMarker
		    (getEdgeValueMarker(),
		     Double.toString(node.getEdgeValue()));
	    }
	    sb.append(getNodeDataTemplate().getPaintedTemplate());
	}
	return sb.toString();	
    }
    
    public void parseIn(Node tree, String treeInfo) {
	throw new FatalException("Tag "+getTagRef()+": This method is not "+
				 "supported by this parser yet.");
    }

}
